"""
Errors raised at instantiation of IntegerEvents and DiceRecord
"""


class InvalidEventsError(ValueError):
    def __init__(self, message='', *args, **kwargs):
        super(InvalidEventsError, self).__init__(message, *args, **kwargs)


class DiceRecordError(ValueError):
    def __init__(self, message='', *args, **kwargs):
        super(DiceRecordError, self).__init__(message, *args, **kwargs)



from __future__ import absolute_import

import unittest

from dicetables.eventsbases.eventerrors import InvalidEventsError, DiceRecordError


class TestEventsErrors(unittest.TestCase):
    def test_InvalidEventsError_empty(self):
        error = InvalidEventsError()
        self.assertEqual(str(error), '')
        self.assertEqual(error.args[0], '')

    def test_InvalidEventsError_non_empty(self):
        error = InvalidEventsError('message')
        self.assertEqual(str(error), 'message')
        self.assertEqual(error.args[0], 'message')

    def test_DiceRecordError_message(self):
        error = DiceRecordError('oops')
        self.assertEqual(error.args[0], 'oops')
        self.assertEqual(str(error), 'oops')

    def test_DiceRecordError_no_message(self):
        error = DiceRecordError()
        self.assertEqual(error.args[0], '')


if __name__ == '__main__':
    unittest.main()



from inspect import signature
from typing import List, Type, Union

import pytest

from dicetables import (
    Die,
    BestOfDicePool,
    ModDie,
    ModWeightedDie,
    Exploding,
    ExplodingOn,
)
from dicetables.dicepool import DicePool
from dicetables.eventsbases.protodie import ProtoDie
from dicetables.tools.limit_checker import NoOpLimitChecker, LimitChecker, LimitsError
from dicetables.tools.orderedcombinations import count_unique_combination_keys


class TestNoOpLimitChecker(object):
    @pytest.fixture(params=[signature(open).bind("x"), signature(divmod).bind(1, 2)])
    def bound_args(self, request):
        yield request.param

    @pytest.fixture(params=[[], [Die], [BestOfDicePool], [Die, BestOfDicePool] * 100])
    def dice(self, request):
        yield request.param

    def test_assert_dice_pool_within_limits(self, bound_args):
        NoOpLimitChecker().assert_dice_pool_within_limits(bound_args)

    def test_assert_die_size_within_limits(self, bound_args):
        NoOpLimitChecker().assert_die_size_within_limits(bound_args)

    def test_assert_numbers_of_calls_within_limits(self, dice):
        NoOpLimitChecker().assert_numbers_of_calls_within_limits(dice)

    def test_assert_explosions_within_limits(self, dice):
        NoOpLimitChecker().assert_explosions_within_limits(dice)


class TestLimitChecker(object):
    def test_limit_checker_defaults(self):
        checker = LimitChecker()
        assert checker.max_dice_calls == 5
        assert checker.max_dice_pool_calls == 2
        assert checker.max_size == 500
        assert checker.max_explosions == 10
        assert checker.max_dice_pool_combinations_per_dict_size == {
            2: 600,
            3: 8700,
            4: 30000,
            5: 55000,
            6: 70000,
            7: 100000,
            12: 200000,
            30: 250000,
        }

    @pytest.mark.parametrize("dice, raises_error", [(2, False), (3, False), (4, True)])
    def test_assert_number_of_calls_within_limits_dice_limit_three(
            self, dice, raises_error
    ):
        checker = LimitChecker(max_dice=3, max_dice_pools=4)
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_numbers_of_calls_within_limits([Die] * dice)
        else:
            checker.assert_numbers_of_calls_within_limits([Die] * dice)

    @pytest.mark.parametrize("dice, raises_error", [(2, False), (3, False), (4, True)])
    def test_assert_number_of_calls_within_limits_max_dice_ignores_dice_pools(
        self, dice, raises_error
    ):
        checker = LimitChecker(max_dice=3, max_dice_pools=4)
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_numbers_of_calls_within_limits([Die, DicePool] * dice)
        else:
            checker.assert_numbers_of_calls_within_limits([Die, DicePool] * dice)

    @pytest.mark.parametrize(
        "dice_pools, raises_error", [(2, False), (3, False), (4, True)]
    )
    def test_assert_number_of_calls_within_limits_dice_pool_limit_three(
        self, dice_pools, raises_error
    ):
        checker = LimitChecker(max_dice=5, max_dice_pools=3)
        to_check = []  # type: List[Union[Type[ProtoDie], Type[DicePool]]]
        to_check += [BestOfDicePool] * 5
        to_check += [DicePool] * dice_pools
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_numbers_of_calls_within_limits(to_check)
        else:
            checker.assert_numbers_of_calls_within_limits(to_check)

    @pytest.mark.parametrize(
        "die_size, raises_error", [(9, False), (10, False), (11, True)]
    )
    def test_assert_die_size_die_with_size_argument(self, die_size, raises_error):
        checker = LimitChecker(max_size=10)
        bound_args = signature(ModDie).bind(die_size, 1000)
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_die_size_within_limits(bound_args)
        else:
            checker.assert_die_size_within_limits(bound_args)

    @pytest.mark.parametrize(
        "die_size, raises_error", [(9, False), (10, False), (11, True)]
    )
    def test_assert_die_size_die_with_dictionary_argument(self, die_size, raises_error):
        checker = LimitChecker(max_size=10)
        bound_args = signature(ModWeightedDie).bind({die_size: 1}, 1000)
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_die_size_within_limits(bound_args)
        else:
            checker.assert_die_size_within_limits(bound_args)

    def test_assert_die_size_ignores_input_dice(self):
        input_die = ModDie(1000, 1000)
        bound_args = signature(Exploding).bind(input_die)
        checker = LimitChecker(max_size=5)
        checker.assert_die_size_within_limits(bound_args)

    def test_assert_explosions_no_explosions(self):
        checker = LimitChecker()
        bound_args = signature(Die).bind(1000)
        checker.assert_explosions_within_limits(bound_args)

    @pytest.mark.parametrize(
        "explosions, raises_error", [(4, False), (5, False), (6, True)]
    )
    def test_assert_explosions_no_explodes_on(self, explosions, raises_error):
        input_die = ModDie(100, 100)
        bound_args = signature(Exploding).bind(input_die, explosions)
        checker = LimitChecker(max_explosions=5)
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_explosions_within_limits(bound_args)
        else:
            checker.assert_explosions_within_limits(bound_args)

    @pytest.mark.parametrize(
        "explodes_on_len, raises_error", [(4, False), (5, False), (6, True)]
    )
    def test_assert_explosions_explosions_plus_len_explodes_on(
        self, explodes_on_len, raises_error
    ):
        input_die = ModDie(100, 100)
        explosions = 10
        explodes_on = tuple(range(explodes_on_len))
        bound_args = signature(ExplodingOn).bind(
            input_die, explodes_on, explosions=explosions
        )
        checker = LimitChecker(max_explosions=5 + explosions)
        if raises_error:
            with pytest.raises(LimitsError):
                checker.assert_explosions_within_limits(bound_args)
        else:
            checker.assert_explosions_within_limits(bound_args)

    def test_assert_dice_pool_limits_no_dice_pool(self):
        checker = LimitChecker()
        bound_args = signature(Die).bind(1000)
        checker.assert_dice_pool_within_limits(bound_args)

    @pytest.mark.parametrize(
        "die_size,die_pool,max_combinations",
        [
            (2, 599, 600),
            (3, 130, 8700),
            (4, 54, 30000),
            (5, 31, 55000),
            (6, 21, 70000),
            (7, 16, 100000),
            (12, 9, 200000),
            (30, 4, 250000),
            (100, 3, 250000),
        ],
    )
    def test_assert_dice_pool_within_limits_on_dice_pool(
        self, die_size, die_pool, max_combinations
    ):
        """
        self.max_dice_pool_combinations_per_dict_size = {
            2: 600, 3: 8700, 4: 30000, 5: 55000, 6: 70000,
            7: 100000, 12: 200000, 30: 250000
        }
        """
        checker = LimitChecker()
        bound_args_passing = signature(DicePool).bind(Die(die_size), die_pool)
        bound_args_failing = signature(DicePool).bind(
            Die(die_size), die_pool + 1
        )
        assert (
            count_unique_combination_keys(Die(die_size), die_pool) <= max_combinations
        )
        checker.assert_dice_pool_within_limits(bound_args_passing)

        assert (
            count_unique_combination_keys(Die(die_size), die_pool + 1)
            > max_combinations
        )
        with pytest.raises(LimitsError):
            checker.assert_dice_pool_within_limits(bound_args_failing)

    def test_assert_dice_pool_within_limits_pool_of_one_fails(self):
        checker = LimitChecker()
        bound_args = signature(DicePool).bind(Die(1), 1)
        with pytest.raises(LimitsError):
            checker.assert_dice_pool_within_limits(bound_args)



import unittest

from dicetables import AdditiveEvents
from dicetables.tools.orderedcombinations import (get_combination_occurrences, ordered_combinations_of_events,
                                                  count_unique_combination_keys, largest_permitted_pool_size
                                                  )


class TestOrderedCombinations(unittest.TestCase):
    def test_count_number_of_combinations_single_key(self):
        to_use = {1: 1, 2: 2, 3: 3, 4: 4}
        self.assertEqual(get_combination_occurrences((1,), to_use), 1)
        self.assertEqual(get_combination_occurrences((2,), to_use), 2)
        self.assertEqual(get_combination_occurrences((3,), to_use), 3)
        self.assertEqual(get_combination_occurrences((4,), to_use), 4)

    def test_count_number_of_combinations_single_key_repeated(self):
        to_use = {1: 1, 2: 2, 3: 3, 4: 4}
        self.assertEqual(get_combination_occurrences((1, 1, 1), to_use), 1)
        self.assertEqual(get_combination_occurrences((2, 2, 2), to_use), 8)
        self.assertEqual(get_combination_occurrences((3, 3, 3), to_use), 27)
        self.assertEqual(get_combination_occurrences((4, 4, 4), to_use), 64)

    def test_count_number_of_combinations_all_values_one_all_possible_groups(self):
        to_use = dict.fromkeys(range(1, 5), 1)
        self.assertEqual(get_combination_occurrences((1, 1, 1), to_use), 1)
        self.assertEqual(get_combination_occurrences((2, 2, 2), to_use), 1)
        self.assertEqual(get_combination_occurrences((3, 3, 3), to_use), 1)
        self.assertEqual(get_combination_occurrences((4, 4, 4), to_use), 1)
        self.assertEqual(get_combination_occurrences((1, 1, 2), to_use), 3)
        self.assertEqual(get_combination_occurrences((1, 1, 3), to_use), 3)
        self.assertEqual(get_combination_occurrences((1, 1, 4), to_use), 3)
        self.assertEqual(get_combination_occurrences((1, 2, 2), to_use), 3)
        self.assertEqual(get_combination_occurrences((1, 3, 3), to_use), 3)
        self.assertEqual(get_combination_occurrences((1, 4, 4), to_use), 3)
        self.assertEqual(get_combination_occurrences((1, 2, 3), to_use), 6)
        self.assertEqual(get_combination_occurrences((1, 2, 4), to_use), 6)
        self.assertEqual(get_combination_occurrences((1, 3, 4), to_use), 6)
        self.assertEqual(get_combination_occurrences((2, 3, 4), to_use), 6)

    def test_count_number_of_combination_all_values_one_commutative_property(self):
        to_use = dict.fromkeys(range(1, 15), 1)
        self.assertEqual(get_combination_occurrences((1, 2, 2, 3, 3, 3), to_use), 60)  # 6!/(1!*2!*3!)
        self.assertEqual(get_combination_occurrences((1, 2, 2, 2, 3, 3), to_use), 60)  # 6!/(1!*3!*2!)
        self.assertEqual(get_combination_occurrences((1, 1, 2, 3, 3, 3), to_use), 60)  # 6!/(2!*1!*3!)
        self.assertEqual(get_combination_occurrences((1, 1, 2, 2, 2, 3), to_use), 60)  # 6!/(2!*3!*1!)
        self.assertEqual(get_combination_occurrences((1, 1, 1, 2, 3, 3), to_use), 60)  # 6!/(3!*1!*2!)
        self.assertEqual(get_combination_occurrences((1, 1, 1, 2, 2, 3), to_use), 60)  # 6!/(3!*2!*1!)

    def test_count_number_of_combinations_all_values_one_different_groupings(self):
        to_use = dict.fromkeys(range(1, 15), 1)
        self.assertEqual(get_combination_occurrences((1, 2), to_use), 2)
        self.assertEqual(get_combination_occurrences((1, 2, 2, 3), to_use), 12)
        self.assertEqual(get_combination_occurrences((1, 1, 2, 2), to_use), 6)
        self.assertEqual(get_combination_occurrences((1, 2, 3, 4), to_use), 24)
        self.assertEqual(get_combination_occurrences((1, 1, 2, 2, 2), to_use), 10)  # 5!/(2!*3!)
        self.assertEqual(get_combination_occurrences((1, 2, 2, 3, 3), to_use), 30)  # 5!/(1!*2!*2!)
        self.assertEqual(get_combination_occurrences((1, 1, 2, 3, 3), to_use), 30)  # 5!/(2!*1!*2!)
        self.assertEqual(get_combination_occurrences((1, 1, 2, 2, 3, 3), to_use), 90)  # 6!/(2!*2!*2!)
        self.assertEqual(get_combination_occurrences((1, 2, 2, 2, 3, 3), to_use), 60)  # 6!/(1!*3!*2!)

    def test_count_number_of_combination_different_values(self):
        to_use = {key: key for key in range(1, 11)}
        self.assertEqual(get_combination_occurrences((1, 2), to_use), 4)
        self.assertEqual(get_combination_occurrences((1, 2, 2, 3), to_use), 12 * 4 * 3)
        self.assertEqual(get_combination_occurrences((2, 2, 3, 3), to_use), 6 * 4 * 9)
        self.assertEqual(get_combination_occurrences((5, 5, 10, 10), to_use), 6 * 25 * 100)

    def test_ordered_combinations_of_events_all_single_weight_simple_case(self):
        answer = ordered_combinations_of_events(AdditiveEvents({1: 1, 2: 1}), 3)
        self.assertEqual(answer, {(1, 1, 1): 1,
                                  (1, 1, 2): 3,
                                  (1, 2, 2): 3,
                                  (2, 2, 2): 1})

    def test_ordered_combinations_of_events_different_weight_is_same_as_combining_integer_events(self):
        answer = ordered_combinations_of_events(AdditiveEvents({1: 1, 2: 2, 3: 3}), 3)
        self.assertEqual(answer, {(1, 1, 1): 1,  # 3: 1
                                  (1, 1, 2): 6,  # 4: 6
                                  (1, 2, 2): 12, (1, 1, 3): 9,  # 5: 12+9
                                  (2, 2, 2): 8, (1, 2, 3): 36,  # 6: 8+36
                                  (1, 3, 3): 27, (2, 2, 3): 36,  # 7: 27+36
                                  (2, 3, 3): 54,  # 8: 54
                                  (3, 3, 3): 27})  # 9: 27
        answer = AdditiveEvents.new().combine(AdditiveEvents({1: 1, 2: 2, 3: 3}), 3).get_dict()
        self.assertEqual(answer, {3: 1, 4: 6, 5: 21, 6: 44, 7: 63, 8: 54, 9: 27})

    def test_count_unique_combination_keys(self):
        events = AdditiveEvents({key: key for key in range(1, 7)})
        for times in range(1, 5):
            number_of_keys = len(ordered_combinations_of_events(events, times).keys())
            self.assertEqual(number_of_keys, count_unique_combination_keys(events, times))

    def test_largest_permitted_pool_size_zero(self):
        events = AdditiveEvents(dict.fromkeys(range(3), 1))
        max_keys = 2
        self.assertEqual(largest_permitted_pool_size(events, max_keys), 0)

    def test_largest_permitted_pool_size_keys_equal_pool_size(self):
        events = AdditiveEvents(dict.fromkeys(range(3), 1))
        pool_size = 5
        self.assertEqual(count_unique_combination_keys(events, pool_size), 21)
        max_keys = 21
        self.assertEqual(largest_permitted_pool_size(events, max_keys), pool_size)

    def test_largest_permitted_pool_size_keys_gt_pool_size(self):
        events = AdditiveEvents(dict.fromkeys(range(3), 1))
        pool_size = 5
        next_pool_size = 6
        self.assertEqual(count_unique_combination_keys(events, pool_size), 21)
        self.assertEqual(count_unique_combination_keys(events, next_pool_size), 28)
        for max_keys in range(22, 28):
            self.assertEqual(largest_permitted_pool_size(events, max_keys), pool_size)
        self.assertEqual(largest_permitted_pool_size(events, 28), next_pool_size)



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods, line-too-long

"""test indexedlist"""
from __future__ import absolute_import

import unittest

from dicetables.tools import indexedvalues as iv


class TestIndexedValues(unittest.TestCase):

    def assert_indexed_values(self, indexed_values, start_index, values):
        self.assertEqual(indexed_values.start_index, start_index)
        self.assertEqual(indexed_values.raw_values, values)

    def test_make_start_index_and_list_start_zero(self):
        start_index, lst = iv.make_start_index_and_list(sorted({0: 1, 1: 1}.items()))
        self.assertEqual(start_index, 0)
        self.assertEqual(lst, [1, 1])

    def test_make_start_index_and_list_start_other(self):
        start_index, lst = iv.make_start_index_and_list(sorted({-2: 1, -1: 1}.items()))
        self.assertEqual(start_index, -2)
        self.assertEqual(lst, [1, 1])

    def test_make_start_index_and_list_gap_in_values(self):
        start_index, lst = iv.make_start_index_and_list(sorted({-2: 1, 1: 1}.items()))
        self.assertEqual(start_index, -2)
        self.assertEqual(lst, [1, 0, 0, 1])

    def test_generated_indexed_values(self):
        test = iv.generate_indexed_values([(1, 1), (2, 2)])
        self.assert_indexed_values(test, 1, [1, 2])

    def test_generated_indexed_values_gap_in_list(self):
        test = iv.generate_indexed_values([(1, 1), (3, 2), (5, 5)])
        self.assert_indexed_values(test, 1, [1, 0, 2, 0, 5])

    def test_generate_indexed_values_from_dict(self):
        test = iv.generate_indexed_values_from_dict({1: 1})
        self.assert_indexed_values(test, 1, [1])

    def test_generate_indexed_values_from_dict_gap_in_values(self):
        test = iv.generate_indexed_values_from_dict({5: 1, 3: 3, 1: 1})
        self.assert_indexed_values(test, 1, [1, 0, 3, 0, 1])

    def test_IndexedValues_init(self):
        test = iv.IndexedValues(1, [1, 3])
        self.assert_indexed_values(test, 1, [1, 3])

    def test_IndexedValues_init_is_mutated_list_safe(self):
        the_list_to_mutate = [1, 2, 3]
        test = iv.IndexedValues(0, the_list_to_mutate)
        the_list_to_mutate[0] = 5
        self.assertEqual(test.raw_values, [1, 2, 3])

    def test_IndexedValues_start_index(self):
        test = iv.IndexedValues(3, [1])
        self.assertEqual(test.start_index, 3)

    def test_IndexedValues_values(self):
        test = iv.IndexedValues(3, [1, 2])
        self.assertEqual(test.raw_values, [1, 2])

    def test_indexedValues_range(self):
        test = iv.IndexedValues(3, [1, 2, 3])
        self.assertEqual(test.index_range, (3, 5))

    def test_IndexedValues_get_dict_single_value(self):
        test = iv.IndexedValues(5, [2])
        self.assertEqual(test.get_dict(), {5: 2})

    def test_IndexedValues_get_dict_does_not_include_zeroes(self):
        test = iv.IndexedValues(1, [2, 0, 3])
        self.assertEqual(test.get_dict(), {1: 2, 3: 3})

    def test_indexedValues_get_value_at_key_within_range(self):
        test = iv.IndexedValues(3, [1, 2, 3])
        self.assertEqual(test.get_value_at_key(4), 2)

    def test_indexedValues_get_value_at_key_out_of_range(self):
        test = iv.IndexedValues(3, [1, 2, 3])
        self.assertEqual(test.get_value_at_key(2), 0)
        self.assertEqual(test.get_value_at_key(6), 0)

    def test_IndexedValues_change_list_len_with_zeroes_max_size_offset_zero(self):
        self.assertEqual(iv.change_list_len_with_zeroes([1, 2, 3], 3, 0), [1, 2, 3])

    def test_change_list_len_with_zeroes_other_case_offset_zero(self):
        self.assertEqual(iv.change_list_len_with_zeroes([1, 2, 3], 5, 0), [1, 2, 3, 0, 0])

    def test_change_list_len_with_zeroes_no_adds(self):
        self.assertEqual(iv.change_list_len_with_zeroes([1, 2, 3], 3, 0), [1, 2, 3])

    def test_change_list_len_with_zeroes_offset_and_zeros_is_total_size(self):
        self.assertEqual(iv.change_list_len_with_zeroes([1, 2, 3], 5, 2), [0, 0, 1, 2, 3])

    def test_change_list_len_with_zeroes_offset_and_zeros_lt_total_size(self):
        self.assertEqual(iv.change_list_len_with_zeroes([1, 2, 3], 10, 2), [0, 0, 1, 2, 3, 0, 0, 0, 0, 0])

    def test_equalize_len_demonstration(self):
        lower = [1, 2, 3, 4, 5, 6]
        higher = [1, 2, 3]
        diff_in_start_indices = 2
        total_size = 6
        self.assertEqual(iv.change_list_len_with_zeroes(lower, total_size, 0),
                         [1, 2, 3, 4, 5, 6])
        self.assertEqual(iv.change_list_len_with_zeroes(higher, total_size, diff_in_start_indices),
                         [0, 0, 1, 2, 3, 0])

    def test_add_many_empty(self):
        self.assertEqual(iv.add_many(), 0)

    def test_add_many_non_empty(self):
        self.assertEqual(iv.add_many(-1, 0, 1, 2, 3), 5)

    def test_get_events_list_normal_case(self):
        occurrences = 3
        new_size = 6
        offset = 2
        event_list = iv.get_events_list([1, 2, 3], occurrences, new_size, offset)
        self.assertEqual(event_list, [0, 0, 3, 6, 9, 0])

    def test_get_events_list_empty(self):
        occurrences = 3
        new_size = 6
        offset = 2
        event_list = iv.get_events_list([], occurrences, new_size, offset)
        self.assertEqual(event_list, [0, 0, 0, 0, 0, 0])

    def test_get_events_list_occurrences_are_one(self):
        occurrences = 1
        new_size = 6
        offset = 2
        event_list = iv.get_events_list([1, 2, 3], occurrences, new_size, offset)
        self.assertEqual(event_list, [0, 0, 1, 2, 3, 0])

    def test_IndexedValues_combine_with_events_list_normal_case(self):
        test = iv.IndexedValues(5, [1, 2, 1])
        input_dict = {1: 2, 2: 1}
        """
        [2, 4, 2, 0] +
        [0, 1, 2, 1]
        """
        self.assert_indexed_values(test.combine_with_dictionary(input_dict), 6, [2, 5, 4, 1])

    def test_IndexedValues_combine_with_events_list_gaps_in_list(self):
        test = iv.IndexedValues(5, [1, 2, 1])
        input_dict = {1: 2, 3: 1}
        """
        [2, 4, 2, 0, 0] +
        [0, 0, 1, 2, 1]
        """
        self.assert_indexed_values(test.combine_with_dictionary(input_dict), 6, [2, 4, 3, 2, 1])

    def test_IndexedValues_combine_with_events_list_negative_events(self):
        test = iv.IndexedValues(5, [1, 2, 1])
        input_dict = {-1: 2, 0: 1}
        """
        [2, 4, 2, 0] +
        [0, 1, 2, 1]
        """
        self.assert_indexed_values(test.combine_with_dictionary(input_dict), 4, [2, 5, 4, 1])


if __name__ == '__main__':
    unittest.main()



"""
All the descendants of ProtoDie.  These are IntegerEvents that represent different types of dice.
"""
import itertools
from typing import Dict, Iterable, Tuple

from dicetables.eventsbases.protodie import ProtoDie


class Modifier(ProtoDie):
    """
    stores and returns info for a modifier to add to the final die roll.
    :code:`Modifier(-3)` rolls -3 and only -3. A Modifier's size and weight are
    always 0.
    """

    def __init__(self, modifier: int):
        self._mod = modifier
        super(Modifier, self).__init__()

    def get_modifier(self) -> int:
        return self._mod

    def get_size(self):
        return 0

    def get_weight(self):
        return 0

    def get_dict(self):
        return {self._mod: 1}

    def weight_info(self):
        return str(self)

    def multiply_str(self, number):
        return '\n'.join([str(self)] * number)

    def __str__(self):
        return '{:+}'.format(self._mod)

    def __repr__(self):
        return 'Modifier({})'.format(self._mod)


class Die(ProtoDie):
    """
    stores and returns info for a basic Die.
    :code:`Die(4)` rolls 1, 2, 3, 4 with equal weight
    """

    def __init__(self, die_size: int):
        """

        :param die_size: int > 0
        """
        self._die_size = die_size
        super(Die, self).__init__()

    def get_size(self):
        return self._die_size

    def get_weight(self):
        return 0

    def get_dict(self) -> Dict[int, int]:
        return dict.fromkeys(range(1, self._die_size + 1), 1)

    def weight_info(self):
        return '{}\n    No weights'.format(self)

    def multiply_str(self, number):
        return '{}{}'.format(number, self)

    def __str__(self):
        return 'D{}'.format(self._die_size)

    def __repr__(self):
        return 'Die({})'.format(self._die_size)


class ModDie(Die):
    """
    stores and returns info for a Die with a modifier
    that changes the values of the rolls.
    :code:`ModDie(4, -1)` rolls 0, 1, 2, 3 with equal weight
    """

    def __init__(self, die_size: int, modifier: int):
        """

        :param die_size: int >0
        :param modifier: int
        """
        self._mod = modifier
        super(ModDie, self).__init__(die_size)

    def get_modifier(self) -> int:
        return self._mod

    def get_dict(self) -> Dict[int, int]:
        return dict.fromkeys(range(1 + self._mod, self.get_size() + 1 + self._mod), 1)

    def multiply_str(self, number):
        return '{}D{}{:+}'.format(number, self.get_size(), number * self._mod)

    def __str__(self):
        return 'D{}{:+}'.format(self.get_size(), self._mod)

    def __repr__(self):
        return 'ModDie({}, {})'.format(self.get_size(), self._mod)


class WeightedDie(ProtoDie):
    """
    stores and returns info for die with different chances for different rolls.
    :code:`WeightedDie({1:1, 2:5})` rolls 1 once for every five times that 2 is rolled.
    """

    def __init__(self, dictionary_input: Dict[int, int]):
        """

        :param dictionary_input: {roll: weight} roll: int>1, weight: int>=0\n
            the sum of all weights >0
        """

        self._raw_dic = self._create_raw_dict(dictionary_input)
        super(WeightedDie, self).__init__()

    def _create_raw_dict(self, dictionary):
        self._raise_value_error_for_rolls_less_than_one(dictionary)
        output = {}
        for key in range(1, max(dictionary.keys()) + 1):
            output[key] = dictionary.get(key, 0)
        return output

    @staticmethod
    def _raise_value_error_for_rolls_less_than_one(dictionary):
        if any(roll < 1 for roll in dictionary):
            raise ValueError('rolls may not be less than 1. use ModWeightedDie')

    def get_raw_dict(self) -> Dict[int, int]:
        return self._raw_dic.copy()

    def get_size(self):
        return max(self._raw_dic.keys())

    def get_weight(self):
        return sum(self._raw_dic.values())

    def get_dict(self) -> Dict[int, int]:
        return {key: value for key, value in self._raw_dic.items() if value}

    def weight_info(self):
        max_roll_str_len = len(str(self.get_size()))
        out = str(self) + '\n'
        for roll, weight in sorted(self.get_raw_dict().items()):
            out += '    a roll of {:>{}} has a weight of {}\n'.format(roll, max_roll_str_len, weight)
        return out.rstrip('\n')

    def multiply_str(self, number):
        return '{}{}'.format(number, self)

    def __str__(self):
        return 'D{}  W:{}'.format(self.get_size(), self.get_weight())

    def __repr__(self):
        return 'WeightedDie({})'.format(self._raw_dic)


class ModWeightedDie(WeightedDie):
    """
    stores and returns info for die with different chances for different rolls.
    The modifier changes all die rolls.
    :code:`ModWeightedDie({1:1, 3:5}, -1)` is a 3-sided die - 1. It
    rolls 0 once for every five times that 2 is rolled.
    """

    def __init__(self, dictionary_input: Dict[int, int], modifier: int):
        """

        :param dictionary_input: {roll: weight} roll: int, weight: int>=0\n
            sum of all weights >0
        :param modifier: int
        """
        self._mod = modifier
        super(ModWeightedDie, self).__init__(dictionary_input)

    def get_modifier(self) -> int:
        return self._mod

    def get_dict(self):
        return {roll + self._mod: weight for roll, weight in self.get_raw_dict().items() if weight}

    def multiply_str(self, number):
        return '{}D{}{:+}  W:{}'.format(number, self.get_size(),
                                        number * self._mod, self.get_weight())

    def __str__(self):
        return 'D{}{:+}  W:{}'.format(self.get_size(), self._mod, self.get_weight())

    def __repr__(self):
        return 'ModWeightedDie({}, {})'.format(self.get_raw_dict(), self._mod)


class StrongDie(ProtoDie):
    """
    stores and returns info for a stronger version of another die (including
    StrongDie if you're feeling especially silly).
    The multiplier multiplies all die rolls of original Die.
    :code:`StrongDie(ModDie(3, -1), 2)` rolls (1-1)*2, (2-1)*2, (3-1)*2 with equal weight.
    """

    def __init__(self, input_die: ProtoDie, multiplier: int):
        """

        :param input_die: Die, ModDie, WeightedDie, ModWeightedDie, StrongDie or subclass of ProtoDie
        :param multiplier: int
        """
        self._multiplier = multiplier
        self._original = input_die
        super(StrongDie, self).__init__()

    def get_size(self):
        return self._original.get_size()

    def get_weight(self):
        return self._original.get_weight()

    def get_multiplier(self):
        return self._multiplier

    def get_input_die(self) -> ProtoDie:
        """returns an instance of the original die"""
        return self._original

    def get_dict(self) -> Dict[int, int]:
        return {roll * self._multiplier: weight for roll, weight in self._original.get_dict().items()}

    def weight_info(self):
        return self._original.weight_info().replace(str(self._original), str(self))

    def multiply_str(self, number):
        return '({})X({})'.format(self._original.multiply_str(number), self._multiplier)

    def __str__(self):
        return '({})X({})'.format(self._original, self._multiplier)

    def __repr__(self):
        return 'StrongDie({!r}, {})'.format(self._original, self._multiplier)


class Exploding(ProtoDie):
    """
    Stores and returns info for an exploding version of another die.
    Each time the highest number is rolled, you
    add that to the total and keep rolling. An exploding D6 rolls 1-5 as usual.
    When it rolls a 6, it re-rolls and adds that 6. If it rolls a 6 again, this continues,
    adding 12 to the result. Since this is an infinite but increasingly unlikely process,
    the "explosions" parameter sets the number of re-rolls allowed.

    Explosions are applied after modifiers and multipliers.
    :code:`Exploding(ModDie(4, -2))` explodes on a 2 so it rolls:
    [-1, 0, 1, (2 -1), (2 + 0), (2 + 1), (2+2 - 1) ..]

    **WARNING:** setting the number of explosions too high can make
    instantiation VERY slow. The time is proportional to explosions and die_size.
    """

    def __init__(self, input_die: ProtoDie, explosions: int = 2):
        """

        :param input_die: Die, ModDie, WeightedDie, ModWeightedDie, StrongDie or subclass of ProtoDie
        :param explosions: int >=0
        """
        self._original = input_die
        self._explosions = explosions
        self._raise_error_for_negative_explosions()
        self._dict = self._get_exploding_dict()
        super(Exploding, self).__init__()

    def _raise_error_for_negative_explosions(self):
        if self._explosions < 0:
            raise ValueError('"explosions" must be >=0.')

    def _get_exploding_dict(self):
        base_dict = self._original.get_dict()

        new_dict = {}
        for explosion_level in range(self._explosions + 1):
            level_dict = self._get_level_dict(explosion_level, base_dict)
            new_dict = add_dicts(new_dict, level_dict)
        return new_dict

    def _get_level_dict(self, explosion_level, base_dict):
        highest_roll = max(base_dict)

        all_occurrences = sum(base_dict.values())
        level_depth_multiplier = all_occurrences ** (self._explosions - explosion_level)
        highest_roll_multiplier = base_dict[highest_roll] ** explosion_level

        occurrence_mod = level_depth_multiplier * highest_roll_multiplier
        roll_mod = explosion_level * highest_roll

        level_dict = {roll + roll_mod: occurrence * occurrence_mod for roll, occurrence in base_dict.items()}
        if explosion_level == self._explosions:
            return level_dict
        return remove_keys_after_applying_modifier(level_dict, (highest_roll,), roll_mod)

    def get_size(self):
        return self._original.get_size()

    def get_weight(self):
        return self._original.get_weight() + 1

    def get_explosions(self) -> int:
        return self._explosions

    def get_input_die(self) -> ProtoDie:
        """returns an instance of the original die"""
        return self._original

    def get_dict(self) -> Dict[int, int]:
        return self._dict.copy()

    def weight_info(self):
        return '{}\nExploding adds weight: 1'.format(
            self._original.weight_info().replace(str(self._original), str(self)))

    def multiply_str(self, number):
        return '{}({})'.format(number, self)

    def __str__(self):
        return '{}: Explosions={}'.format(self._original, self._explosions)

    def __repr__(self):
        return 'Exploding({!r}, {})'.format(self._original, self._explosions)


class ExplodingOn(ProtoDie):
    """
    Stores and returns info for an exploding version of another die.
    Each time the values in (explodes_on) are rolled, the die continues to roll,
    adding that value to the result. The die only continues rolling an (explosions) number of times.

    :code:`ExplodingOn(Die(6), (1, 6), explosions=2)` rolls:
    [2 to 5], 1+[2 to 5], 6+[2 to 5], 1+1+[1 to 6], 1+6+[1 to 6], 6+1+[1 to 6] and 6+6+[1 to 6].

    Explosions are applied after modifiers and multipliers.
    :code:`ExplodingOn(ModDie(4, -2), (2,))` explodes on a 2 so it rolls:
    [-1, 0, 1, (2 -1), (2 + 0), (2 + 1), (2+2 - 1) ..]

    **WARNING:** setting the number of explosions too high can make
    instantiation VERY slow. Time is proportional to explosion**(len(explodes_on)). It's also linear
    with size which gets overshadowed by the first factor.
    """

    def __init__(self, input_die: ProtoDie, explodes_on: Iterable[int], explosions: int=2):
        """

        :param input_die: Die, ModDie, WeightedDie, ModWeightedDie, StrongDie or subclass of ProtoDie
        :param explodes_on: tuple[int] - rolls that die explodes on. Must be in keys of input_die.get_dict().
        :param explosions: int >=0
        """
        self._explodes_on = remove_duplicates(explodes_on)
        self._explosions = explosions
        self._original = input_die
        self._raise_error_for_bad_explodes_on()
        self._raise_error_for_negative_explosions()
        self._dict = self._get_exploding_dict()
        super(ExplodingOn, self).__init__()

    def _raise_error_for_bad_explodes_on(self):
        base_dict = self._original.get_dict()
        if any(key not in base_dict for key in self._explodes_on):
            raise ValueError('"explodes_on" value not present in input_die.get_dict()')

    def _raise_error_for_negative_explosions(self):
        if self._explosions < 0:
            raise ValueError('"explosions" value must be >= 0.')

    def _get_exploding_dict(self):
        base_dict = self._original.get_dict()
        answer = {}
        for level in range(self._explosions + 1):
            roll_and_weight_modifiers = self._get_roll_and_weight_mods(level, base_dict)
            level_dict = self._get_level_dict(base_dict, level, roll_and_weight_modifiers)
            answer = add_dicts(answer, level_dict)
        return answer

    def _get_roll_and_weight_mods(self, level, base_dict):
        if level == 0:
            return [(0, 1)]
        base_roll_weight_mods = [(roll, base_dict[roll]) for roll in self._explodes_on]
        groups_of_roll_weight_mods = itertools.product(base_roll_weight_mods, repeat=level)
        return [calc_roll_and_weight_mods(group_of_rollweights) for group_of_rollweights in groups_of_roll_weight_mods]

    def _get_level_dict(self, base_dict, level, roll_weights):
        answer = {}
        roll_weight_dict = combine_rollweights_with_same_roll_value(roll_weights)
        for roll_mod, weight_mod in roll_weight_dict.items():
            to_add = self._get_partial_level_dict(base_dict, level, roll_mod, weight_mod)
            answer = add_dicts(answer, to_add)
        return answer

    def _get_partial_level_dict(self, base_dict, level, roll_mod, weight_multiplier):
        base_level_multiplier = sum(base_dict.values())
        level_multiplier = base_level_multiplier ** (self._explosions - level)
        current_level_all_rolls = {roll + roll_mod: occurrence * weight_multiplier * level_multiplier
                                   for roll, occurrence in base_dict.items()}
        if level == self._explosions:
            return current_level_all_rolls

        return remove_keys_after_applying_modifier(current_level_all_rolls, self._explodes_on, roll_mod)

    def get_size(self):
        return self._original.get_size()

    def get_weight(self):
        return self._original.get_weight() + len(self._explodes_on)

    def get_explosions(self):
        return self._explosions

    def get_explodes_on(self) -> Tuple[int, ...]:
        return self._explodes_on

    def get_input_die(self) -> ProtoDie:
        """returns an instance of the original die"""
        return self._original

    def get_dict(self) -> Dict[int, int]:
        return self._dict.copy()

    def weight_info(self):
        base_weight_info = self._original.weight_info().replace(str(self._original), str(self))
        num_of_values = len(self._explodes_on)
        value_str = 'value' if num_of_values == 1 else 'values'
        return '{0}\nExploding on {1} {2} adds weight: {1}'.format(base_weight_info, num_of_values, value_str)

    def multiply_str(self, number):
        return '{}({})'.format(number, self)

    def __str__(self):
        explodes_on = ', '.join([str(val) for val in self._explodes_on])
        return '{}: Explosions={} On: {}'.format(self._original, self._explosions, explodes_on)

    def __repr__(self):
        return 'ExplodingOn({!r}, {}, {})'.format(self._original, self._explodes_on, self._explosions)


def remove_duplicates(input_tuple: Iterable[int]) -> Tuple[int, ...]:
    list_version = []
    for val in input_tuple:
        if val not in list_version:
            list_version.append(val)
    return tuple(list_version)


def add_dicts(dict_1, dict_2):
    out = dict_1.copy()
    for key, val in dict_2.items():
        out[key] = out.get(key, 0) + val
    return out


def calc_roll_and_weight_mods(group_of_rollweights):
    total_roll_mod = 0
    total_weight_mod = 1
    for roll, weight in group_of_rollweights:
        total_roll_mod += roll
        total_weight_mod *= weight
    return total_roll_mod, total_weight_mod


def remove_keys_after_applying_modifier(modified_dict, original_excluded_keys, key_modifier):
    return {key: val for key, val in modified_dict.items() if key - key_modifier not in original_excluded_keys}


def combine_rollweights_with_same_roll_value(rollweight_tuples):
    answer = {}
    for roll, weight in rollweight_tuples:
        answer[roll] = weight + answer.get(roll, 0)
    return answer



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import

import unittest
from sys import version_info

from dicetables.eventsbases.eventerrors import InvalidEventsError
from dicetables.eventsbases.integerevents import IntegerEvents, EventsVerifier


class DummyEvents(IntegerEvents):
    def __init__(self, dictionary):
        self._dictionary = dictionary
        super(DummyEvents, self).__init__()

    def get_dict(self):
        return self._dictionary


class TestIntegerEvents(unittest.TestCase):
    def setUp(self):
        self.checker = EventsVerifier()
        self.types_error = 'all values must be ints'
        if version_info[0] < 3:
            self.types_error += ' or longs'

    def tearDown(self):
        del self.checker
        del self.types_error

    def assert_my_regex(self, error_type, regex, func, *args):
        with self.assertRaises(error_type) as cm:
            func(*args)
        error_msg = cm.exception.args[0]
        self.assertEqual(error_msg, regex)

    def test_assert_my_regex(self):
        self.assert_my_regex(ValueError, "invalid literal for int() with base 10: 'a'", int, 'a')

    def test_EventsVerifier_verify_get_dict_pass(self):
        self.assertIsNone(self.checker.verify_get_dict({1: 1}))

    def test_EventsVerifier_verify_get_dict_empty(self):
        self.assert_my_regex(InvalidEventsError,
                             'events may not be empty. a good alternative is the identity - {0: 1}.',
                             self.checker.verify_get_dict, {})

    def test_EventsVerifier_verify_get_dict_zero_occurrences(self):
        self.assert_my_regex(InvalidEventsError,
                             'no negative or zero occurrences in Events.get_dict()',
                             self.checker.verify_get_dict, {1: 0, 2: 0})

    def test_EventsVerifier_verify_get_dict_negative_occurrences(self):
        self.assert_my_regex(InvalidEventsError, 'no negative or zero occurrences in Events.get_dict()',
                             self.checker.verify_get_dict, {1: -1})

    def test_EventsVerifier_verify_get_dict_non_int_occurrences(self):
        self.assert_my_regex(InvalidEventsError, self.types_error,
                             self.checker.verify_get_dict, {1: 1.0})

    def test_EventsVerifier_verify_get_dict_non_int_event(self):
        self.assert_my_regex(InvalidEventsError, self.types_error,
                             self.checker.verify_get_dict, {1.0: 1})

    def test_EventsVerifier_is_all_ints_pass(self):
        self.assertTrue(self.checker.is_all_ints([10 ** value for value in range(500)]))

    def test_EventsVerifier_all_ints_fail(self):
        self.assertFalse(self.checker.is_all_ints([1.0, 1, 1, 1, 1, 1]))

    def test_EventsVerifier_does_not_work_if_does_not_follow_minimum_requirements(self):
        self.assertRaises(AttributeError, self.checker.verify_get_dict, 'a')
        self.assertRaises(AttributeError, self.checker.verify_get_dict, [1, 2, 3])
        self.assertRaises(AttributeError, self.checker.verify_get_dict, [(1, 2, 3), (4, 5, 6)])
        self.assertRaises(InvalidEventsError, self.checker.verify_get_dict, {'a': 'b'})

    def test_IntegerEvents_checks_get_dict_at_init(self):
        self.assertRaises(InvalidEventsError, DummyEvents, {'a': 'b'})

    def test_IntegerEvents_init_error_message(self):
        message = ('get_dict() must return a dictionary\n' +
                   '{event: occurrences, ...} event=int, occurrence=int>0.')
        self.assert_my_regex(NotImplementedError, message, IntegerEvents)

    def test_IntegerEvents__eq__true(self):
        self.assertTrue(DummyEvents({1: 2}).__eq__(DummyEvents({1: 2})))

    def test_IntegerEvents__eq__false_by_type_of_Events(self):
        class NewDummy(DummyEvents):
            pass

        self.assertFalse(DummyEvents({1: 2}).__eq__(NewDummy({1: 2})))
        self.assertFalse(NewDummy({1: 2}).__eq__(DummyEvents({1: 2})))

    def test_IntegerEvents__eq__false_by_unrelated_object(self):
        self.assertFalse(DummyEvents({1: 2}).__eq__(2))

    def test_IntegerEvents__eq__false_by_get_dict(self):
        self.assertFalse(DummyEvents({1: 2}).__eq__(DummyEvents({1: 2, 3: 4})))

    def test_IntegerEvents__ne__false(self):
        self.assertFalse(DummyEvents({1: 2}).__ne__(DummyEvents({1: 2})))

    def test_IntegerEvents__ne__true_by_type_of_Events(self):
        class NewDummy(DummyEvents):
            pass

        self.assertTrue(DummyEvents({1: 2}).__ne__(NewDummy({1: 2})))
        self.assertTrue(NewDummy({1: 2}).__ne__(DummyEvents({1: 2})))

    def test_IntegerEvents__ne__true_by_unrelated_object(self):
        self.assertTrue(DummyEvents({1: 2}).__ne__(2))

    def test_IntegerEvents__ne__true_by_get_dict(self):
        self.assertTrue(DummyEvents({1: 2}).__ne__(DummyEvents({1: 2, 3: 4})))


if __name__ == '__main__':
    unittest.main()



"""
sole purpose: the top function - get_string_from_list_of_ints
"""


def get_string_from_list_of_ints(input_list):
    to_use = sorted(input_list)
    list_of_sequences = split_at_gaps_larger_than_one(to_use)
    list_of_sequence_strings = format_sequences(list_of_sequences)
    return ', '.join(list_of_sequence_strings)


def split_at_gaps_larger_than_one(sorted_list):
    list_of_sequences = []
    for value in sorted_list:
        if not list_of_sequences or gap_is_larger_than_one(list_of_sequences, value):
            list_of_sequences.append([value])
        else:
            last_group = list_of_sequences[-1]
            last_group.append(value)
    return list_of_sequences


def gap_is_larger_than_one(list_of_sequences, value):
    max_gap_size = 1
    last_value_of_list = list_of_sequences[-1][-1]
    return value - last_value_of_list > max_gap_size


def format_sequences(list_of_sequences):
    string_list = []
    for sequence in list_of_sequences:
        string_list.append(format_one_sequence(sequence))
    return string_list


def format_one_sequence(sequence):
    first = sequence[0]
    last = sequence[-1]
    if first == last:
        return format_for_sequence_str(first)
    else:
        return '{}-{}'.format(format_for_sequence_str(first), format_for_sequence_str(last))


def format_for_sequence_str(num):
    return '({:,})'.format(num) if num < 0 else '{:,}'.format(num)



from itertools import combinations_with_replacement
from math import factorial
from typing import Dict, Tuple

from dicetables.eventsbases.integerevents import IntegerEvents


def count_unique_combination_keys(events: IntegerEvents, pool_size: int) -> int:
    """calculates how large the set of keys will be for `ordered_combinations_of_events(events, pool_size)`."""
    ans = 1
    dict_size = len(events.get_dict())
    for number in range(dict_size, dict_size + pool_size):
        ans *= number
    return ans // factorial(pool_size)


def largest_permitted_pool_size(events, max_number_of_keys):
    pool_size = 0
    num_keys = 0
    while num_keys <= max_number_of_keys:
        pool_size += 1
        num_keys = count_unique_combination_keys(events, pool_size)
    return pool_size - 1


def ordered_combinations_of_events(events: IntegerEvents, times: int) -> Dict[Tuple[int, ...], int]:
    base_dict = events.get_dict()
    ordered_combinations = combinations_with_replacement(sorted(base_dict.keys()), times)
    return {key: get_combination_occurrences(key, base_dict) for key in ordered_combinations}


def get_combination_occurrences(combination, base_dict):
    """simple calculation over tuple of sorted values and possibly repeating value
    to tell number of unique permutations.  so get_count((1, 2, 3)) = 3!.
    get_count((1, 2, 2, 3, 3)) = 5!/(2!*2!)"""
    answer = factorial(len(combination))
    current = combination[0]
    multiplier = base_dict[current]
    count = 0
    for el in combination:
        if el != current:
            count = 0
            multiplier = base_dict[el]
            current = el

        count += 1
        answer //= count
        answer *= multiplier

    return answer



"""
DictCombiner combines dictionary of {event value: occurrences} and returns a DictCombiner.  Retrieve the result with
the "get_dict()" method.

THIS IS A SPECIFIC TOOL. IT HAS SEVERAL IMPORTANT CONSTRAINTS. USE WITH CAUTION.

if you mutate the __init__ dictionary elsewhere, IT WILL AFFECT IT HERE. (but it will
not mutate any dictionaries passed to it.)
    this class is for speed, so there is minimal copying. it is specifically
    designed to be instantiated once, used and then immediately discarded.

:__init__ dictionary: {int: int>0, ...}
:variable - times: int >=0
:variable - dictionary: {int: int>0, ...}
"""
from dicetables.tools.indexedvalues import generate_indexed_values_from_dict


class DictCombiner(object):
    def __init__(self, dictionary: dict):
        """
        :dictionary: {int: int>0, ...}
        """
        self._dict = dictionary

    def get_dict(self) -> dict:
        return self._dict.copy()

    def combine_by_flattened_list(self, dictionary: dict, times: int) -> dict:
        """
        :dictionary: {int: int>0, ...}
        """
        new_combiner = DictCombiner(self.get_dict())
        flattened_list = flatten_events_tuples(dictionary)
        for _ in range(times):
            new_combiner = new_combiner.combine_once_with_flattened_list(flattened_list)
        return new_combiner.get_dict()

    def combine_once_with_flattened_list(self, flattened_list) -> 'DictCombiner':
        new_dict = {}
        for event, current_frequency in self._dict.items():
            for new_event in flattened_list:
                new_dict[event + new_event] = (new_dict.get(event + new_event, 0) + current_frequency)
        return DictCombiner(new_dict)

    def combine_by_dictionary(self, dictionary: dict, times: int) -> dict:
        """
        :dictionary: {int: int>0, ...}
        """
        new_combiner = DictCombiner(self.get_dict())
        for _ in range(times):
            new_combiner = new_combiner.combine_once_with_dictionary(dictionary)
        return new_combiner.get_dict()

    def combine_once_with_dictionary(self, dictionary):
        new_dict = {}
        for event, current_frequency in self._dict.items():
            for new_event, frequency in dictionary.items():
                new_dict[event + new_event] = (new_dict.get(event + new_event, 0) + frequency * current_frequency)
        return DictCombiner(new_dict)

    def combine_by_indexed_values(self, dictionary: dict, times: int) -> dict:
        """
        :dictionary: {int: int>0, ...}
        """
        new_indexed_values = generate_indexed_values_from_dict(self._dict)
        for _ in range(times):
            new_indexed_values = new_indexed_values.combine_with_dictionary(dictionary)
        return new_indexed_values.get_dict()

    def combine_by_fastest(self, dictionary: dict, times: int) -> dict:
        """
        :dictionary: {int: int>0, ...}
        """
        method_dict = {'dictionary': self.combine_by_dictionary,
                       'flattened_list': self.combine_by_flattened_list,
                       'indexed_values': self.combine_by_indexed_values}
        method = self.get_fastest_combine_method(dictionary, times)

        return method_dict[method](dictionary, times)

    def get_fastest_combine_method(self, dictionary, times):
        """
        :dictionary: {int: int>0, ...}
        """
        first_comparison = self._compare_tuple_list_with_flattened_list(dictionary)
        second_comparison = self._compare_with_indexed_values(first_comparison, times, dictionary)
        return second_comparison

    @staticmethod
    def _compare_tuple_list_with_flattened_list(dictionary):
        max_occurrences_to_events_ratio_for_flattened_list = 1.3
        safe_limit_flattened_list_len = 10 ** 4
        total_occurrences = sum(dictionary.values())

        if total_occurrences >= safe_limit_flattened_list_len:
            return 'dictionary'

        occurrences_to_events_ratio = float(total_occurrences) / len(dictionary)
        if occurrences_to_events_ratio > max_occurrences_to_events_ratio_for_flattened_list:
            return 'dictionary'
        return 'flattened_list'

    def _compare_with_indexed_values(self, first_method, times, dictionary):
        size_of_dict_to_combine = len(dictionary)
        min_size_for_indexed_values = get_indexed_values_min(first_method, size_of_dict_to_combine, times)
        size_of_main_dict = len(self._dict)
        if size_of_main_dict < min_size_for_indexed_values:
            return first_method
        else:
            return 'indexed_values'

    def remove_by_tuple_list(self, dictionary, times):
        """
        :dictionary: {int: int>0, ...}
        """
        events_tuples = sorted(dictionary.items())
        new_events = DictCombiner(self.get_dict())
        for _ in range(times):
            new_events = new_events.remove_once_by_tuple_list(events_tuples)
        return new_events.get_dict()

    def remove_once_by_tuple_list(self, events_tuples):
        new_dict_max, new_dict_min = self.get_new_dict_range(events_tuples)
        new_dict = {}
        for target_event in range(new_dict_min, new_dict_max + 1):

            freq_at_new_event = self.get_target_event_freq(target_event, events_tuples, new_dict)
            if freq_at_new_event:
                new_dict[target_event] = freq_at_new_event

        return DictCombiner(new_dict)

    def get_new_dict_range(self, events_tuples):
        min_event_being_removed = events_tuples[0][0]
        max_event_being_removed = events_tuples[-1][0]
        current_min_event = min(self._dict.keys())
        current_max_event = max(self._dict.keys())

        new_dict_min = current_min_event - min_event_being_removed
        new_dict_max = current_max_event - max_event_being_removed
        return new_dict_max, new_dict_min

    def get_target_event_freq(self, target_event, events_tuples, new_dict):
        min_event_being_removed = events_tuples[0][0]
        try:
            old_freq = self._dict[target_event + min_event_being_removed]
        except KeyError:
            return 0
        freq_at_new_event = old_freq
        for event_being_removed, event_weight in events_tuples[1:]:
            removal_value_offset = event_being_removed - min_event_being_removed
            freq_at_new_event -= new_dict.get(target_event - removal_value_offset, 0) * event_weight
        freq_at_new_event = freq_at_new_event // events_tuples[0][1]
        return freq_at_new_event


def flatten_events_tuples(dictionary):
    flattened_list = []
    for event, freq in dictionary.items():
        flattened_list += [event] * freq
    return flattened_list


def get_indexed_values_min(first_method, size_of_dict_to_combine, combine_times):
    """
    {'first method': {size of input dict: {combine times: size of Dictcombiner.get_dict(), ...}, ...}, ... }

    :first_method: 'flattened_list', 'dictionary'
    """
    choices = {
        'flattened_list':
            {
                2: {10: 200, 20: 100, 100: 50, 500: 1},
                4: {2: 100, 5: 50, 10: 10, 20: 1},
                6: {1: 100, 4: 50, 10: 1},
                8: {1: 100, 3: 50, 5: 10, 10: 1},
                20: {1: 50, 3: 10, 4: 1},
                50: {1: 50, 3: 1},
                100: {1: 100, 2: 1}
            },
        'dictionary':
            {
                2: {10: 100, 50: 50, 100: 1},
                4: {2: 100, 10: 50, 50: 1},
                6: {2: 50, 10: 10, 20: 1},
                8: {1: 100, 3: 50, 10: 1},
                10: {1: 100, 3: 50, 5: 1},
                20: {1: 50, 3: 10, 4: 1},
                50: {1: 50, 2: 10, 3: 1},
                100: {1: 100, 2: 1}
            }
    }
    keys_are_input_dict_size = choices[first_method]

    closest_dict_size = get_best_key(size_of_dict_to_combine, keys_are_input_dict_size)
    keys_are_times = keys_are_input_dict_size[closest_dict_size]
    closest_times = get_best_key(combine_times, keys_are_times)
    min_size_for_indexed_values = keys_are_times[closest_times]
    return min_size_for_indexed_values


def get_best_key(value, dictionary):
    sorted_keys = sorted(dictionary.keys())
    best_key = sorted_keys[0]
    for test in sorted_keys:
        if value < test:
            break
        best_key = test
    return best_key



from collections import namedtuple

Alias = namedtuple('Alias', ['primary', 'alternate', 'primary_height'])


class AliasTable(object):
    """

    here is
    `a nice explanation of alias tables: <http://www.keithschwarz.com/darts-dice-coins/>`_

    `Vose's algorithm <https://web.archive.org/web/20131029203736/http://web.eecs.utk.edu/~vose/Publications/random.pdf>`_

    """

    def __init__(self, input_dict):
        self._height = sum(input_dict.values())
        self._length = len(input_dict)
        self._aliases = self._create_aliases(input_dict)

    def _create_aliases(self, input_dict):
        big_heights, small_heights = self._get_height_sorted_lists(input_dict)

        alias_list = []
        while small_heights:
            primary, primary_height = small_heights.pop()
            alternate, alternate_height = big_heights.pop()
            alias_list.append(Alias(primary=primary, alternate=alternate, primary_height=primary_height))
            new_alternate_height = alternate_height - (self._height - primary_height)
            self._update_sorting_lists(alternate, new_alternate_height, big_heights, small_heights)

        while big_heights:
            primary, _ = big_heights.pop()
            alias_list.append(Alias(primary=primary, alternate=primary, primary_height=self._height))

        return alias_list

    def _update_sorting_lists(self, event, event_height, big_heights, small_heights):
        if event_height < self._height:
            small_heights.append((event, event_height))
        else:
            big_heights.append((event, event_height))

    def _get_height_sorted_lists(self, input_dict):
        less_than_height = []
        greater_than_or_equal_height = []
        for event, frequency in sorted(input_dict.items()):
            event_height = self._length * frequency
            self._update_sorting_lists(event, event_height, greater_than_or_equal_height, less_than_height)
        return greater_than_or_equal_height, less_than_height

    @property
    def height(self):
        return self._height

    @property
    def length(self):
        return self._length

    def to_list(self):
        return self._aliases[:]

    def get(self, length, height):
        alias = self._aliases[length]
        if height >= alias.primary_height:
            return alias.alternate
        return alias.primary



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import

import unittest

from dicetables.dicetable import DiceTable
from dicetables.dieevents import Die
from dicetables.factory.errorhandler import EventsFactoryError, EventsFactoryErrorHandler
from dicetables.factory.eventsfactory import EventsFactory, Getter


class StandInForBadClass(object):
    pass


class TestEventsFactoryErrorHandler(unittest.TestCase):
    def setUp(self):
        EventsFactory.reset()

    def assert_events_factory_error_message(self, msg, func, *args):
        with self.assertRaises(EventsFactoryError) as cm:
            func(*args)
        error_msg = cm.exception.args[0]
        self.assertEqual(error_msg, msg)

    def test_assert_events_factory_error_message(self):
        def my_func():
            raise EventsFactoryError('hello')

        self.assert_events_factory_error_message('hello', my_func)

    def test_EventsFactoryErrorHandler_CLASS_OVERWRITE(self):
        expected = (
                'Error Code: CLASS OVERWRITE\n' +
                'Factory:    <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Error At:   <class \'dicetables.dicetable.DiceTable\'>\n' +
                '\n' +
                'Attempted to add class already in factory but used different factory keys.\n' +
                'Class: <class \'dicetables.dicetable.DiceTable\'>\n' +
                'Current Factory Keys: (\'get_dict\', \'dice_data\')\n' +
                'Keys Passed In:       (\'dice_data\',)\n'
        )
        self.assert_events_factory_error_message(expected,
                                                 EventsFactoryErrorHandler(EventsFactory).raise_error,
                                                 'CLASS OVERWRITE',
                                                 DiceTable,
                                                 ('dice_data',))

    def test_EventsFactoryErrorHandler_GETTER_OVERWRITE(self):
        expected = (
                'Error Code: GETTER OVERWRITE\n' +
                'Factory:    <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Error At:   Factory Getter Key: \'dice_data\'\n' +
                '\n' +
                'Attempted to add getter key already in factory but used different parameters.\n' +
                'Key: \'dice_data\'\n' +
                'Factory Parameter:    method: "dice_data", default: DiceRecord({})\n' +
                'Passed In Parameters: method: "get_dice", default: [(2, Die(6))]\n'
        )
        self.assert_events_factory_error_message(expected,
                                                 EventsFactoryErrorHandler(EventsFactory).raise_error,
                                                 'GETTER OVERWRITE',
                                                 'dice_data',
                                                 Getter('get_dice', [(2, Die(6))]))

    def test_EventsFactoryErrorHandler_MISSING_GETTER(self):
        expected = (
                'Error Code: MISSING GETTER\n' +
                'Factory:    <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Error At:   <class \'tests.factory.test_factoryerrorhandler.StandInForBadClass\'>\n' +
                '\n' +
                'Attempted to add class with a getter key not in the factory.\n' +
                'Class: <class \'tests.factory.test_factoryerrorhandler.StandInForBadClass\'>\n' +
                'Current Factory Keys: [\'calc_includes_zeroes\', \'dice_data\', \'get_dict\']\n' +
                'Key Passed In:        \'foo\'\n'
        )
        self.assert_events_factory_error_message(expected,
                                                 EventsFactoryErrorHandler(EventsFactory).raise_error,
                                                 'MISSING GETTER',
                                                 StandInForBadClass,
                                                 'foo')

    def test_EventsFactoryErrorHandler_SIGNATURES_DIFFERENT(self):
        expected = (
                'Error Code: SIGNATURES DIFFERENT\n' +
                'Factory:    <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Error At:   <class \'dicetables.dicetable.DiceTable\'>\n' +
                '\n' +
                'Attempted to construct a class already present in factory, but with a different signature.\n' +
                'Class: <class \'dicetables.dicetable.DiceTable\'>\n' +
                'Signature In Factory: (\'get_dict\', \'dice_data\')\n' +
                'To reset the factory to its base state, use EventsFactory.reset()\n'
        )
        self.assert_events_factory_error_message(expected,
                                                 EventsFactoryErrorHandler(EventsFactory).raise_error,
                                                 'SIGNATURES DIFFERENT',
                                                 DiceTable)

    def test_EventsFactoryErrorHandler_WTF(self):
        expected = (
                'Error Code: WTF\n' +
                'Factory:    <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Error At:   <class \'tests.factory.test_factoryerrorhandler.StandInForBadClass\'>\n' +
                '\n' +
                'Attempted to construct a class unrelated, in any way, to any class in the factory.\n' +
                'EventsFactory can currently construct the following classes:\n' +
                '[\'AdditiveEvents\', \'DetailedDiceTable\', \'DiceTable\']\n' +
                'EventsFactory searched the MRO of <class \'tests.factory.test_factoryerrorhandler.StandInForBadClass\'>,\n'
                + 'and found no matches to the classes in the factory.\n'
        )
        self.assert_events_factory_error_message(expected,
                                                 EventsFactoryErrorHandler(EventsFactory).raise_error,
                                                 'WTF',
                                                 StandInForBadClass)


if __name__ == '__main__':
    unittest.main()



"""
The abstract class for sets of events that can be represented by integers.
"""
from typing import Dict

from dicetables.eventsbases.eventerrors import InvalidEventsError


class EventsVerifier(object):

    def verify_get_dict(self, events_dict):
        """

        :param events_dict: IntegerEvents.get_dict()
        :raises: InvalidEventsError
        """
        if not events_dict:
            raise InvalidEventsError('events may not be empty. a good alternative is the identity - {0: 1}.')
        if not self.is_all_ints(events_dict.keys()) or not self.is_all_ints(events_dict.values()):
            raise InvalidEventsError('all values must be ints')
        if any(occurrence <= 0 for occurrence in events_dict.values()):
            raise InvalidEventsError('no negative or zero occurrences in Events.get_dict()')

    @staticmethod
    def is_all_ints(iterable):
        return all(isinstance(value, int) for value in iterable)


class IntegerEvents(object):
    def __init__(self):
        super(IntegerEvents, self).__init__()
        EventsVerifier().verify_get_dict(self.get_dict())

    def get_dict(self) -> Dict[int, int]:
        """

        :return: {event: occurrences}
        """
        raise NotImplementedError(('get_dict() must return a dictionary\n'
                                   '{event: occurrences, ...} event=int, occurrence=int>0.'))

    def __eq__(self, other):
        return type(self) is type(other) and self.get_dict() == other.get_dict()

    def __ne__(self, other):
        return not self == other



"""
An immutable record of dice added to and removed from DiceTable
"""
from typing import Dict

from dicetables.eventsbases.eventerrors import DiceRecordError
from dicetables.eventsbases.protodie import ProtoDie


class RecordVerifier(object):
    @staticmethod
    def check_types(die_input):
        if any(not isinstance(die, ProtoDie) or not isinstance(num, int) for die, num in die_input.items()):
            raise DiceRecordError('input must be {ProtoDie: int, ...}')

    @staticmethod
    def check_negative(die_input):
        for key, val in die_input.items():
            if val < 0:
                raise DiceRecordError('Tried to create a DiceRecord with a negative value at {!r}: {}'.format(key, val))

    @staticmethod
    def check_number(num):
        if num < 0:
            raise DiceRecordError('Tried to add_die or remove_die with a negative number.')


def scrub_zeroes(input_dict):
    return {key: val for key, val in input_dict.items() if val}


class DiceRecord(object):
    def __init__(self, dice_number_dict: Dict[ProtoDie, int]):
        RecordVerifier.check_types(dice_number_dict)
        RecordVerifier.check_negative(dice_number_dict)
        self._record = scrub_zeroes(dice_number_dict)

    @classmethod
    def new(cls) -> 'DiceRecord':
        return cls({})

    def get_dict(self) -> Dict[ProtoDie, int]:
        return self._record.copy()

    def get_number(self, query_die: ProtoDie) -> int:
        return self._record.get(query_die, 0)

    def add_die(self, die: ProtoDie, times: int) -> 'DiceRecord':
        RecordVerifier.check_number(times)
        new = self._record.copy()
        new[die] = times + self.get_number(die)
        return DiceRecord(new)

    def remove_die(self, die: ProtoDie, times: int) -> 'DiceRecord':
        RecordVerifier.check_number(times)
        new = self._record.copy()
        new[die] = self.get_number(die) - times
        return DiceRecord(new)

    def __eq__(self, other):
        if not isinstance(other, DiceRecord):
            return False
        return self.get_dict() == other.get_dict()

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return 'DiceRecord({!r})'.format(self._record)



"""
For getting and computing details for any IntegerEvents, including:

- points/axes for graphing, both raw and pct
    - pct can be computed quickly up to 10 decimal places or more exactly and slowly using Decimal
    - can include all the zero values between the lowest and highest non-zero events in an IntegerEvents
- stddev and mean
- string of all numbers in events in human readable form
- strings for the percent chance of any subset of events within an events

Can be accessed through objects or wrapper functions.
"""

from collections import namedtuple
from decimal import Decimal
from math import log10
from typing import List, Tuple

from dicetables.eventsbases.integerevents import IntegerEvents
from dicetables.tools.listtostring import get_string_from_list_of_ints
from dicetables.tools.numberforamtter import NumberFormatter


def safe_true_div(numerator, denominator):
    """floating point division for any sized number"""
    ans = Decimal(numerator) / Decimal(denominator)
    return _convert_decimal_to_float_or_int(ans)


def _convert_decimal_to_float_or_int(num):
    answer_as_float = float(num)
    if answer_as_float == float('inf') or answer_as_float == float('-inf'):
        return int(num)
    else:
        return answer_as_float


class EventsInformation(object):
    def __init__(self, events: IntegerEvents):
        self._dict = events.get_dict()

    def get_items(self):
        """

        :return: dict.items(): a list in py2 and an iterator in py3.
        """
        return self._dict.items()

    def events_keys(self) -> List[int]:
        return sorted(self._dict.keys())

    def events_range(self) -> Tuple[int, int]:
        all_keys = self.events_keys()
        return all_keys[0], all_keys[-1]

    def total_occurrences(self) -> int:
        return sum(self._dict.values())

    def all_events(self):
        return sorted(self._dict.items())

    def all_events_include_zeroes(self):
        start, stop = self.events_range()
        return self.get_range_of_events(start, stop + 1)

    def biggest_event(self) -> Tuple[int, int]:
        """

        :return: (event, occurrences) for first event with highest occurrences
        """
        highest_occurrences = max(self._dict.values())
        for event, occurrences in sorted(self._dict.items()):  # pragma: no branch
            if occurrences == highest_occurrences:
                return event, highest_occurrences

    def biggest_events_all(self) -> List[Tuple[int, int]]:
        """

        :return: the list of all events that have biggest occurrence
        """
        highest_occurrences = max(self._dict.values())
        output = []
        for event, occurrences in self._dict.items():
            if occurrences == highest_occurrences:
                output.append((event, occurrences))
        return sorted(output)

    def get_event(self, event: int) -> Tuple[int, int]:
        return event, self._dict.get(event, 0)

    def get_range_of_events(self, start: int, stop_before: int) -> List[Tuple[int, int]]:
        return [self.get_event(event) for event in range(start, stop_before)]


class EventsCalculations(object):

    def __init__(self, events: IntegerEvents, include_zeroes: bool = True):
        self._info = EventsInformation(events)
        self._include_zeroes = include_zeroes

    @property
    def include_zeroes(self) -> int:
        return self._include_zeroes

    @property
    def info(self) -> EventsInformation:
        return self._info

    def mean(self) -> float:
        numerator = sum((value * freq) for value, freq in self._info.get_items())
        denominator = self._info.total_occurrences()
        return safe_true_div(numerator, denominator)

    def stddev(self, decimal_place=4) -> float:
        avg = self.mean()
        factor_to_truncate_digits = self._get_truncation_factor(decimal_place)
        truncated_deviations = 0
        for event_value, occurrences in self._info.get_items():
            truncated_deviations += (occurrences // factor_to_truncate_digits) * (avg - event_value) ** 2.
        truncated_total_occurrences = self._info.total_occurrences() // factor_to_truncate_digits
        return round((truncated_deviations / truncated_total_occurrences) ** 0.5, decimal_place)

    def _get_truncation_factor(self, decimal_place):
        extra_digits = 5
        largest_exponent = int(log10(self._info.biggest_event()[1]))
        required_exp_for_accuracy = 2 * (extra_digits + decimal_place)
        if largest_exponent < required_exp_for_accuracy:
            factor_to_truncate_digits = 1
        else:
            factor_to_truncate_digits = 10 ** (largest_exponent - required_exp_for_accuracy)
        return factor_to_truncate_digits

    def percentage_points(self) -> List[Tuple[int, float]]:
        """
        Very fast, but only good to ten decimal places.
        """
        return self._percentage_points_by_method('fast')

    def percentage_points_exact(self) -> List[Tuple[int, float]]:
        return self._percentage_points_by_method('exact')

    def percentage_axes(self):
        """
        Very fast, but only good to ten decimal places.
        """
        return list(zip(*self.percentage_points()))

    def percentage_axes_exact(self):
        return list(zip(*self.percentage_points_exact()))

    def log10_points(self, log10_of_zero_value=-100.0) -> List[Tuple[int, float]]:
        """
        returns log10 of the occurrences.

        :param log10_of_zero_value: any zero-occurrence must have a preset value.
        """
        return [(event, log10(occurrence) if occurrence != 0 else log10_of_zero_value)
                for event, occurrence in self._get_data_set()]

    def log10_axes(self, log10_of_zero_value=-100.0):
        """
        returns log10 of the occurrences.

        :param log10_of_zero_value: any zero-occurrence must have a preset value.
        """
        return list(zip(*self.log10_points(log10_of_zero_value)))

    def _percentage_points_by_method(self, method_str):
        methods = {'fast': get_fast_pct_number, 'exact': get_exact_pct_number}
        pct_method = methods[method_str]
        total_values = self._info.total_occurrences()
        return [(event, pct_method(occurrence, total_values)) for event, occurrence in self._get_data_set()]

    def _get_data_set(self):
        if self._include_zeroes:
            return self._info.all_events_include_zeroes()
        else:
            return self._info.all_events()

    def full_table_string(self, shown_digits=4, max_power_for_commaed=6):
        """

        :param shown_digits: How many digits in each scientific notation string.
        :param max_power_for_commaed: The largest power to be represented in comma notation.
            if set to -1, all numbers are in scientific notation.
        """
        shown_digits = max(shown_digits, 1)
        formatter = NumberFormatter(shown_digits=shown_digits, max_comma_exp=max_power_for_commaed)
        right_just = self._get_right_just()
        out_str = ''
        for value, frequency in self._get_data_set():
            out_str += '{:>{}}: {}\n'.format(value, right_just, formatter.format(frequency))
        return out_str

    def _get_right_just(self):
        min_event, max_event = self._info.events_range()
        value_right_just = max(len(str(min_event)), len(str(max_event)))
        return value_right_just

    def stats_strings(self, query_list, shown_digits=4, max_power_for_commaed=6, min_power_for_fixed_pt=-3):
        """
        Calculates the pct chance and one-in chance of any list of numbers,
        including numbers not in the Events.

        :param query_list: A list of ints. Calculates the chance of the list getting rolled.
        :param shown_digits: How many digits in each scientific notation number str.
        :param max_power_for_commaed: The largest power to be represented in comma notation.
            if set to -1, all numbers >= 1 are in scientific notation.
        :param min_power_for_fixed_pt: The smallest power to be represented in fixed point notation.
            If set to zero, all values < 1 represented in scientific notation.
        :return: (query values, query occurrences, total occurrences, inverse chance, pct chance)
        """
        shown_digits = max(shown_digits, 1)
        formatter = NumberFormatter(shown_digits=shown_digits, max_comma_exp=max_power_for_commaed,
                                    min_fixed_pt_exp=min_power_for_fixed_pt)

        total_occurrences = self._info.total_occurrences()
        query_values_occurrences = self._get_query_values_occurrences(query_list)

        inverse_chance_str, pct_str = _calculate_chance_and_pct(query_values_occurrences, total_occurrences)

        return StatsStrings(get_string_from_list_of_ints(query_list),
                            formatter.format(query_values_occurrences),
                            formatter.format(total_occurrences),
                            formatter.format(inverse_chance_str),
                            formatter.format(pct_str))

    def _get_query_values_occurrences(self, query_values):
        combinations_of_values = 0
        no_copies = set(query_values)
        for value in no_copies:
            combinations_of_values += self._info.get_event(value)[1]
        return combinations_of_values


StatsStrings = namedtuple('StatsStrings',
                          ['query_values', 'query_occurrences', 'total_occurrences', 'one_in_chance', 'pct_chance'])


def _calculate_chance_and_pct(query_values_occurrences, total_combinations):
    if not query_values_occurrences:
        return Decimal('inf'), Decimal('0')
    inverse_chance = Decimal(total_combinations) / Decimal(query_values_occurrences)
    pct = Decimal(100.0) / inverse_chance
    return inverse_chance, pct


def get_fast_pct_number(number, total_values):
    factor = 10 ** 50
    will_not_overflow = (number * factor) // total_values
    return will_not_overflow * 100. / float(factor)


def get_exact_pct_number(number, total_values):
    return safe_true_div(100 * number, total_values)


# wrappers functions

def events_range(events):
    return EventsInformation(events).events_range()


def mean(events):
    return EventsCalculations(events).mean()


def stddev(events, decimal_place=4):
    return EventsCalculations(events).stddev(decimal_place=decimal_place)


def percentage_points(events, include_zeroes=True):
    return EventsCalculations(events, include_zeroes).percentage_points()


def percentage_axes(events, include_zeroes=True):
    return EventsCalculations(events, include_zeroes).percentage_axes()


def stats(events, query_values, shown_digits=4, max_power_for_commaed=6, min_power_for_fixed_pt=-3):
    return EventsCalculations(events).stats_strings(query_values,
                                                    shown_digits=shown_digits,
                                                    max_power_for_commaed=max_power_for_commaed,
                                                    min_power_for_fixed_pt=min_power_for_fixed_pt)


def full_table_string(events, include_zeroes=True, shown_digits=4, max_power_for_commaed=6):
    return EventsCalculations(events, include_zeroes).full_table_string(shown_digits=shown_digits,
                                                                        max_power_for_commaed=max_power_for_commaed)



from __future__ import absolute_import

import unittest

import dicetables.tools.listtostring as l2s


class TestListToString(unittest.TestCase):
    def test_format_for_sequence_string_adds_commas(self):
        self.assertEqual(l2s.format_for_sequence_str(10), '10')
        self.assertEqual(l2s.format_for_sequence_str(1000000), '1,000,000')

    def test_format_for_sequence_string_negative_numbers(self):
        self.assertEqual(l2s.format_for_sequence_str(-2), '(-2)')
        self.assertEqual(l2s.format_for_sequence_str(-2000), '(-2,000)')

    def test_format_one_sequence_single_number(self):
        self.assertEqual(l2s.format_one_sequence([1]), '1')

    def test_format_one_sequence_first_and_last_same_number(self):
        self.assertEqual(l2s.format_one_sequence([1, 2, 3, 1]), '1')

    def test_format_one_sequence_first_and_last_different(self):
        self.assertEqual(l2s.format_one_sequence([-1, 3, 2]), '(-1)-2')

    def test_gap_is_larger_than_one_true(self):
        self.assertTrue(l2s.gap_is_larger_than_one([[1, 2, 3], [4, 5]], 7))

    def test_gap_is_larger_than_one_false(self):
        self.assertFalse(l2s.gap_is_larger_than_one([[1, 2, 3], [4, 5]], 5))
        self.assertFalse(l2s.gap_is_larger_than_one([[1, 2, 3], [4, 5]], 6))

    def test_split_at_gaps_larger_than_one_empty_list(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([]), [])

    def test_split_at_gaps_larger_than_one_one_element(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([1]), [[1]])

    def test_split_at_gaps_larger_than_one_same_element(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([1, 1, 1, 1]), [[1, 1, 1, 1]])

    def test_split_at_gaps_larger_than_one_not_large_gaps(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([1, 1, 2, 2, 3]), [[1, 1, 2, 2, 3]])

    def test_split_at_gaps_larger_than_one_all_large_gaps(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([1, 3, 9]), [[1], [3], [9]])

    def test_split_at_gaps_larger_than_one_mixed_case(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([1, 1, 2, 5, 6, 9]), [[1, 1, 2], [5, 6], [9]])

    def test_split_at_gaps_larger_than_one_mixed_case_negative_pos_zero(self):
        self.assertEqual(l2s.split_at_gaps_larger_than_one([-2, 0, 1, 2, 5, 6, 9]), [[-2], [0, 1, 2], [5, 6], [9]])

    def test_get_string_from_list_returns_single_number(self):
        self.assertEqual(l2s.get_string_from_list_of_ints([1, 1, 1]), '1')

    def test_get_string_from_list_puts_parentheses_around_negative_numbers(self):
        self.assertEqual(l2s.get_string_from_list_of_ints([-1]), '(-1)')

    def test_get_string_from_list_returns_proper_output_for_a_run_of_numbers(self):
        self.assertEqual(l2s.get_string_from_list_of_ints([1, 1, 2, 3, 4, 5]), '1-5')

    def test_get_string_from_list_returns_values_separated_by_commas(self):
        the_list = [-5, -4, -3, -1, 0, 1, 2, 3, 5]
        self.assertEqual(l2s.get_string_from_list_of_ints(the_list), '(-5)-(-3), (-1)-3, 5')

    def test_get_string_from_list_returns_numbers_with_commas(self):
        the_list = [-1234567, 1234567]
        self.assertEqual(l2s.get_string_from_list_of_ints(the_list), '(-1,234,567), 1,234,567')

    def test_get_string_from_list_returns_number_out_of_order(self):
        the_list = [5, 3, 7, 1, 3, 6, 8, 9, 10, 2]
        self.assertEqual(l2s.get_string_from_list_of_ints(the_list), '1-3, 5-10')


if __name__ == '__main__':
    unittest.main()



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods, line-too-long

from __future__ import absolute_import

import unittest

from dicetables.tools.dictcombiner import flatten_events_tuples, get_best_key, get_indexed_values_min, DictCombiner


class TestDictCombiner(unittest.TestCase):
    def setUp(self):
        self.combiner_size_of_one = DictCombiner({1: 1})

    def tearDown(self):
        del self.combiner_size_of_one

    """
    the next several tests show how DictCombiner.get_fastest_method works.  The edge cases to choose between
    methods were determined empirically.  To see how they were derived see time_trials/flattenedlist_timetrials.py
    and time_trials/indexedvalues_timetrials.py.  Those modules require numpy and matplotlib where the actual
    project does not.  Before running either one, under "if __name__ == '__main__':" make sure to comment in
    the UI you wish to use.  The full data set for determining whether to use indexed_values is this:
        choices = {'flattened_list': {
        2: {10: 200,  20: 100,  100: 50,  500: 1},
        4: {2: 100, 5: 50, 10: 10, 20: 1},
        6: {1: 100, 4: 50, 10: 1},
        8: {1: 100, 3: 50, 5: 10, 10: 1},
        20: {1: 50, 3: 10, 4: 1},
        50: {1: 50, 3: 1},
        100: {1: 100, 2: 1}
        },
        'dictionary': {
        2: {10: 100, 50: 50, 100: 1},
        4: {2: 100, 10: 50, 50: 1},
        6: {2: 50, 10: 10, 20: 1},
        8: {1: 100, 3: 50, 10: 1},
        10: {1: 100, 3: 50, 5: 1},
        20: {1: 50, 3: 10, 4: 1},
        50: {1: 50, 2: 10, 3: 1},
        100: {1: 100, 2: 1}
        }
    }
    {'first method': {size of input dict: {combine times: size of Dictcombiner.get_dict(), ...}, ...}, ... }
    """

    def test_DictCombiner_get_fastest_method_DictCombiner_sized_one_and_one_times_never_picks_indexed_values(self):
        accepted_choices = ('dictionary', 'flattened_list')
        for power_of_two in range(10):
            """
            input_dict is:
            len(input_dict) = 2**power_of_two (start =1, end = about 10**3)
            single = 1 occurrence
            some = 1, 2, 3, 1, 2, 3, ...  occurrences
            high = 1 + 2 ** event value
            """
            single_occurrence, some_occurrence, high_occurrence = next(input_dict_generator())
            self.assertIn(self.combiner_size_of_one.get_fastest_combine_method(single_occurrence, 1), accepted_choices)
            self.assertIn(self.combiner_size_of_one.get_fastest_combine_method(some_occurrence, 1), accepted_choices)
            self.assertIn(self.combiner_size_of_one.get_fastest_combine_method(high_occurrence, 1), accepted_choices)

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_total_occurrences_min(self):
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method({1: 1}, 1), 'flattened_list')

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_total_occurrences_mid(self):
        input_dict = dict.fromkeys(range(100), 1)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(input_dict, 1), 'flattened_list')

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_total_occurrences_edge(self):
        input_dict = dict.fromkeys(range(9999), 1)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(input_dict, 1), 'flattened_list')

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_total_occurrences_over_edge(self):
        """
        this cutoff is not for speed but for safety.  as the next test will demonstrate
        """
        input_dict = dict.fromkeys(range(10000), 1)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(input_dict, 1), 'dictionary')

    def test_DictCombiner_demonstrate_why_there_is_cutoff_for_flattened_list(self):
        ok_events = {1: 10 ** 4}
        self.assertEqual(flatten_events_tuples(ok_events), [1] * 10 ** 4)
        bad_events = {1: 10 ** 20}
        self.assertRaises((OverflowError, MemoryError), flatten_events_tuples, bad_events)

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_ratio_under(self):
        input_dict = {1: 1, 2: 1, 3: 2, 4: 1}
        ratio = sum(input_dict.values()) / float(len(list(input_dict.keys())))
        self.assertEqual(ratio, 1.25)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(input_dict, 1), 'flattened_list')

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_ratio_edge(self):
        input_dict = {0: 1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1, 7: 1, 8: 1, 10: 4}
        ratio = sum(input_dict.values()) / float(len(list(input_dict.keys())))
        self.assertEqual(ratio, 1.3)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(input_dict, 1), 'flattened_list')

    def test_DictCombiner_get_fastest_method_tuple_vs_flattened_by_ratio_over_edge(self):
        input_dict = dict.fromkeys(range(99), 1)
        input_dict[100] = 32
        ratio = sum(input_dict.values()) / float(len(list(input_dict.keys())))
        self.assertEqual(ratio, 1.31)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(input_dict, 1), 'dictionary')

    def test_DictCombiner_get_fastest_method_part_get_best_key_below_min(self):
        test_dict = {1: 1, 3: 1, 5: 1}
        self.assertEqual(get_best_key(-1, test_dict), 1)

    def test_DictCombiner_get_fastest_method_part_get_best_key_above_max(self):
        test_dict = {1: 1, 3: 1, 5: 1}
        self.assertEqual(get_best_key(1001, test_dict), 5)

    def test_DictCombiner_get_fastest_method_part_get_best_key_at_value(self):
        test_dict = {1: 1, 3: 1, 5: 1}
        self.assertEqual(get_best_key(3, test_dict), 3)

    def test_DictCombiner_get_fastest_method_part_get_best_key_between_values(self):
        test_dict = {1: 1, 3: 1, 5: 1}
        self.assertEqual(get_best_key(4, test_dict), 3)

    def test_DictCombiner_get_fastest_method_part_get_indexed_values_min_method_variable(self):
        """
        'flattened_list' - 4: {2: 100, 5: 50, 10: 10, 20: 1}
        'dictionary' - 4: {2: 100, 10: 50, 50: 1}
        """
        self.assertEqual(get_indexed_values_min('flattened_list', 4, 10), 10)
        self.assertEqual(get_indexed_values_min('dictionary', 4, 10), 50)

    def test_DictCombiner_get_fastest_method_part_get_indexed_values_min_uses_get_best_key(self):
        """
        'dictionary' - {2: {10: 100, 50: 50, 100: 1},
                        4: {2: 100, 10: 50, 50: 1}. ...
        """
        self.assertEqual(get_indexed_values_min('dictionary', 4, 10), 50)
        self.assertEqual(get_indexed_values_min('dictionary', 4, 49), 50)
        self.assertEqual(get_indexed_values_min('dictionary', 5, 10), 50)
        self.assertEqual(get_indexed_values_min('dictionary', 1, 1), 100)
        self.assertEqual(get_indexed_values_min('dictionary', 100000, 10000), 1)

    def test_white_box_test_get_indexed_values_min_hardcoded_without_error_because_i_am_a_dumb_ass(self):
        range_of_input_dictionary = [1, 2, 4, 6, 8, 10, 20, 50, 100, 200]
        range_of_combine_times = [1, 2, 3, 4, 5, 10, 20, 50, 100, 200, 500, 600]
        for size in range_of_input_dictionary:
            for times in range_of_combine_times:
                self.assertIsInstance(get_indexed_values_min('dictionary', size, times), int)
                self.assertIsInstance(get_indexed_values_min('flattened_list', size, times), int)

    def test_DictCombiner_get_fastest_method_uses_size_cutoff_to_choose_indexed_values__combiner_size_one(self):
        """'flattened_list' - 4: {2: 100, 5: 50, 10: 10, 20: 1}"""
        sized_twenty = dict.fromkeys(range(20), 1)
        sized_one_combiner = DictCombiner({1: 1})
        self.assertEqual(get_indexed_values_min('flattened_list', 4, 20), 1)
        self.assertEqual(sized_one_combiner.get_fastest_combine_method(sized_twenty, 4),
                         'indexed_values')

    def test_DictCombiner_get_fastest_method_uses_size_cutoff_to_choose_other_method__combiner_size_one(self):
        """'dictionary' - 4: {2: 100, 10: 50, 50: 1}"""
        sized_four = dict.fromkeys(range(4), 2)
        sized_one_combiner = DictCombiner({1: 1})
        self.assertEqual(get_indexed_values_min('dictionary', 4, 20), 50)
        self.assertEqual(sized_one_combiner.get_fastest_combine_method(sized_four, 4), 'dictionary')

    def test_DictCombiner_get_fastest_method_uses_size_cutoff_to_choose__dict_combiner_is_below_cutoff(self):
        """'dictionary' - 4: {2: 100, 10: 50, 50: 1}"""
        sized_four = dict.fromkeys(range(4), 2)
        sized_five_combiner = DictCombiner(dict.fromkeys(range(5), 1))
        self.assertEqual(get_indexed_values_min('dictionary', 4, 10), 50)
        self.assertEqual(sized_five_combiner.get_fastest_combine_method(sized_four, 20),
                         'dictionary')

    def test_DictCombiner_get_fastest_method_uses_size_cutoff_to_choose__dict_combiner_is_at_cutoff(self):
        """'dictionary' - 4: {2: 100, 10: 50, 50: 1}"""
        sized_four = dict.fromkeys(range(4), 2)
        sized_fifty_combiner = DictCombiner(dict.fromkeys(range(50), 1))
        self.assertEqual(get_indexed_values_min('dictionary', 4, 10), 50)
        self.assertEqual(sized_fifty_combiner.get_fastest_combine_method(sized_four, 20),
                         'indexed_values')

    def test_DictCombiner_get_fastest_method_uses_size_cutoff_to_choose__dict_combiner_is_above_cutoff(self):
        """'dictionary' - 4: {2: 100, 10: 50, 50: 1}"""
        sized_four = dict.fromkeys(range(4), 2)
        sized_fifty_one_combiner = DictCombiner(dict.fromkeys(range(51), 1))
        self.assertEqual(get_indexed_values_min('dictionary', 4, 10), 50)
        self.assertEqual(sized_fifty_one_combiner.get_fastest_combine_method(sized_four, 20),
                         'indexed_values')

    def test_DictCombiner_combine_by_dictionary_identity(self):
        to_combine = {1: 1, 2: 2}
        new_dict = DictCombiner({0: 1}).combine_by_dictionary(to_combine, 1)
        self.assertEqual(new_dict, to_combine)

    def test_DictCombiner_combine_by_dictionary_many_combines(self):
        to_combine = {1: 1, 2: 2}
        new_dict = DictCombiner({0: 1}).combine_by_dictionary(to_combine, 3)
        """
        {1: 1, 2: 2}
    
        {2: 1, 3: 2} + {3: 2, 4: 4} = {2: 1, 3: 4, 4: 4}
    
        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        self.assertEqual(new_dict, {3: 1, 4: 6, 5: 12, 6: 8})

    def test_DictCombiner_combine_by_dictionary_input_dict_has_spaces(self):
        to_combine = {10: 1, 20: 2}
        new_dict = DictCombiner({0: 1}).combine_by_dictionary(to_combine, 3)
        """
        {10: 1, 20: 2}

        {20: 1, 30: 2} + {30: 2, 40: 4} = {20: 1, 30: 4, 40: 4}

        {30: 1, 40: 4, 50: 4} + {40: 2, 50: 8, 60:8} = {30:1, 40: 6, 50: 12, 60: 8}
        """
        self.assertEqual(new_dict, {30: 1, 40: 6, 50: 12, 60: 8})

    def test_DictCombiner_combine_by_dictionary_complex_DictCombiner(self):
        to_combine = {1: 1, 2: 2}
        complex_events = DictCombiner({2: 1, 3: 4, 4: 4})
        new_dict = complex_events.combine_by_dictionary(to_combine, 1)
        """
        {2: 1, 3: 4, 4: 4}
    
        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3: 1, 4: 6, 5: 12, 6: 8}
        """
        self.assertEqual(new_dict, {3: 1, 4: 6, 5: 12, 6: 8})

    def test_DictCombiner_combine_by_flattened_list_identity(self):
        to_combine = {1: 1, 2: 2}
        new_dict = DictCombiner({0: 1}).combine_by_flattened_list(to_combine, 1)
        self.assertEqual(new_dict, to_combine)

    def test_DictCombiner_combine_by_flattened_list__many_combines(self):
        to_combine = {1: 1, 2: 2}
        new_dict = DictCombiner({0: 1}).combine_by_flattened_list(to_combine, 3)
        """
        {1: 1, 2: 2}
    
        {2: 1, 3: 2} + {3: 2, 4: 4} = {2:1, 3: 4, 4: 4}
    
        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        self.assertEqual(new_dict, {3: 1, 4: 6, 5: 12, 6: 8})

    def test_DictCombiner_combine_by_flattened_list_input_dict_has_spaces(self):
        to_combine = {10: 1, 20: 2}
        new_dict = DictCombiner({0: 1}).combine_by_flattened_list(to_combine, 3)
        """
        {10: 1, 20: 2}

        {20: 1, 30: 2} + {30: 2, 40: 4} = {20: 1, 30: 4, 40: 4}

        {30: 1, 40: 4, 50: 4} + {40: 2, 50: 8, 60:8} = {30:1, 40: 6, 50: 12, 60: 8}
        """
        self.assertEqual(new_dict, {30: 1, 40: 6, 50: 12, 60: 8})

    def test_DictCombiner_combine_by_flattened_list_complex_DictCombiner(self):
        to_combine = {1: 1, 2: 2}
        complex_events = DictCombiner({2: 1, 3: 4, 4: 4})
        new_dict = complex_events.combine_by_flattened_list(to_combine, 1)
        """
        {2: 1, 3: 4, 4: 4}
    
        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        self.assertEqual(new_dict, {3: 1, 4: 6, 5: 12, 6: 8})

    def test_DictCombiner_combine_by_indexed_values_identity(self):
        to_combine = {1: 1, 2: 2}
        new_dict = DictCombiner({0: 1}).combine_by_indexed_values(to_combine, 1)
        self.assertEqual(new_dict, to_combine)

    def test_DictCombiner_combine_by_indexed_values_many_combines(self):
        to_combine = {1: 1, 2: 2}
        new_dict = DictCombiner({0: 1}).combine_by_indexed_values(to_combine, 3)
        """
        {1: 1, 2: 2}
    
        {2: 1, 3: 2} + {3: 2, 4: 4} = {2:1, 3: 4, 4: 4}
    
        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        self.assertEqual(new_dict, {3: 1, 4: 6, 5: 12, 6: 8})

    def test_DictCombiner_combine_by_indexed_values_input_dict_has_spaces(self):
        to_combine = {10: 1, 20: 2}
        new_dict = DictCombiner({0: 1}).combine_by_indexed_values(to_combine, 3)
        """
        {10: 1, 20: 2}

        {20: 1, 30: 2} + {30: 2, 40: 4} = {20: 1, 30: 4, 40: 4}

        {30: 1, 40: 4, 50: 4} + {40: 2, 50: 8, 60:8} = {30:1, 40: 6, 50: 12, 60: 8}
        """
        self.assertEqual(new_dict, {30: 1, 40: 6, 50: 12, 60: 8})

    def test_DictCombiner_combine_by_indexed_values_complex_DictCombiner(self):
        to_combine = {1: 1, 2: 2}
        complex_events = DictCombiner({2: 1, 3: 4, 4: 4})
        new_dict = complex_events.combine_by_indexed_values(to_combine, 1)
        """
        {2: 1, 3: 4, 4: 4}
    
        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        self.assertEqual(new_dict, {3: 1, 4: 6, 5: 12, 6: 8})

    def test_DictCombiner_combine_by_fastest_works_with_flattened_list(self):
        to_add = {1: 1, 2: 1}
        identity = DictCombiner({0: 1})
        self.assertEqual(identity.get_fastest_combine_method(to_add, 1), 'flattened_list')
        self.assertEqual(identity.combine_by_fastest(to_add, 1), to_add)

    def test_DictCombiner_combine_by_fastest_works_with_dictionary(self):
        to_add = {1: 10, 2: 10}
        identity = DictCombiner({0: 1})
        self.assertEqual(identity.get_fastest_combine_method(to_add, 1), 'dictionary')
        self.assertEqual(identity.combine_by_fastest(to_add, 1), to_add)

    def test_DictCombiner_combine_by_fastest_works_with_indexed_values(self):
        to_add = dict.fromkeys(range(100), 1)
        answer = self.combiner_size_of_one.combine_by_flattened_list(to_add, 2)
        self.assertEqual(self.combiner_size_of_one.get_fastest_combine_method(to_add, 2), 'indexed_values')
        self.assertEqual(self.combiner_size_of_one.combine_by_fastest(to_add, 2), answer)

    def test_DictCombiner_remove_by_tuple_list(self):
        """
        {1: 1, 2: 2}

        {2: 1, 3: 2} + {3: 2, 4: 4} = {2: 1, 3: 4, 4: 4}

        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        start = DictCombiner({3: 1, 4: 6, 5: 12, 6: 8})
        new = start.remove_by_tuple_list({1: 1, 2: 2}, 1)
        self.assertEqual(new, {2: 1, 3: 4, 4: 4})

    def test_DictCombiner_remove_by_tuple_list_many_removes(self):
        """
        {1: 1, 2: 2}

        {2: 1, 3: 2} + {3: 2, 4: 4} = {2: 1, 3: 4, 4: 4}

        {3: 1, 4: 4, 5: 4} + {4: 2, 5: 8, 6:8} = {3:1, 4: 6, 5: 12, 6: 8}
        """
        start = DictCombiner({3: 1, 4: 6, 5: 12, 6: 8})
        new = start.remove_by_tuple_list({1: 1, 2: 2}, 3)
        self.assertEqual(new, {0: 1})

    def test_DictCombiner_remove_by_tuple_list_dict_has_spaces_zeroes_not_included(self):
        """
        {10: 1, 20: 2}

        {20: 1, 30: 2} + {30: 2, 40: 4} = {20: 1, 30: 4, 40: 4}

        {30: 1, 40: 4, 50: 4} + {40: 2, 50: 8, 60:8} = {30: 1, 40: 6, 50: 12, 60: 8}
        """
        start = DictCombiner({30: 1, 40: 6, 50: 12, 60: 8})
        new = start.remove_by_tuple_list({10: 1, 20: 2}, 2)
        self.assertEqual(new, {10: 1, 20: 2})


def input_dict_generator():
    step_by_power_of_two = 1
    while True:
        single_occurrence = dict([(event, 1) for event in range(step_by_power_of_two)])
        middle_occurrences = dict([(event, 1 + event % 3) for event in range(step_by_power_of_two)])
        high_occurrences = dict([(event, 1 + 2 * event) for event in range(step_by_power_of_two)])
        step_by_power_of_two *= 2
        yield (single_occurrence,
               middle_occurrences,
               high_occurrences)


if __name__ == '__main__':
    unittest.main()



import unittest

from dicetables.tools.alias_table import AliasTable, Alias


class TestAliasTable(unittest.TestCase):
    def test_alias(self):
        primary = 1
        alternate = 2
        primary_height = 3
        alias = Alias(primary=primary, alternate=alternate, primary_height=primary_height)
        self.assertEqual(alias.primary, primary)
        self.assertEqual(alias.alternate, alternate)
        self.assertEqual(alias.primary_height, primary_height)

    def assert_alias_table_has_expected_counts(self, alias_table, expected_counts):
        actual_counts = dict.fromkeys(expected_counts.keys(), 0)
        for length_val in range(alias_table.length):
            for height_val in range(alias_table.height):
                actual_counts[alias_table.get(length_val, height_val)] += 1

        self.assertEqual(expected_counts, actual_counts)

        actual_primaries = {alias.primary for alias in alias_table.to_list()}
        expected_primaries = set(expected_counts.keys())
        self.assertEqual(actual_primaries, expected_primaries)

    def test_alias_table(self):
        input_dict = {1: 1, 3: 4, 5: 2}
        expected_height = 1 + 4 + 2
        expected_length = 3

        alias_table = AliasTable(input_dict)

        self.assertEqual(alias_table.height, expected_height)
        self.assertEqual(alias_table.length, expected_length)

        expected_counts = {1: 3, 3: 12, 5: 6}

        self.assert_alias_table_has_expected_counts(alias_table, expected_counts)

    def test_even_alias_table_even_number(self):
        input_dict = {1: 2, 2: 2, 3: 2, 4: 2}
        expected_counts = {1: 8, 2: 8, 3: 8, 4: 8}

        alias_table = AliasTable(input_dict)
        self.assertEqual(alias_table.length, 4)
        self.assertEqual(alias_table.height, 8)

        self.assert_alias_table_has_expected_counts(alias_table, expected_counts)

    def test_even_alias_table_odd_number(self):
        input_dict = {1: 2, 2: 2, 4: 2}
        expected_counts = {1: 6, 2: 6, 4: 6}

        alias_table = AliasTable(input_dict)
        self.assertEqual(alias_table.length, 3)
        self.assertEqual(alias_table.height, 6)
        self.assert_alias_table_has_expected_counts(alias_table, expected_counts)

    def test_alias_table_to_list_flat_chances(self):
        input_dict = {1:1, 2: 1, 3: 1}
        alias_table = AliasTable(input_dict)
        result = alias_table.to_list()
        expected = [
            Alias(primary=3, alternate=3, primary_height=3),
            Alias(primary=2, alternate=2, primary_height=3),
            Alias(primary=1, alternate=1, primary_height=3)
        ]
        self.assertEqual(result, expected)

    def test_alias_table_to_list(self):
        input_dict = {1: 1, 2: 2, 3: 2, 4: 3}
        alias_table = AliasTable(input_dict)
        expected_list = [
            Alias(primary=1, alternate=4, primary_height=4),
            Alias(primary=4, alternate=4, primary_height=8),
            Alias(primary=3, alternate=3, primary_height=8),
            Alias(primary=2, alternate=2, primary_height=8)
        ]
        self.assertEqual(alias_table.to_list(), expected_list)

    def test_alias_table_to_list_silly_test_so_that_i_can_visualize_Vose_algorithm_better(self):
        input_dict = {1: 1, 2: 3, 3: 4, 4: 6, 5: 7}
        alias_table = AliasTable(input_dict)
        self.assertEqual(alias_table.height, 21)
        self.assert_alias_table_has_expected_counts(alias_table, {1: 5, 2: 15, 3: 20, 4: 30, 5: 35})
        # small = [(1, 5), (2, 15), (3, 20)]
        # big = [(4, 30), (5, 35)]
        expected_list = [
            Alias(primary=3, alternate=5, primary_height=20),  # small[(1, 5), (2, 15)], big[(4, 30), (5, 35-1)]
            Alias(primary=2, alternate=5, primary_height=15),  # small[(1, 5)], big[(4, 30), (5, 34-6)]
            Alias(primary=1, alternate=5, primary_height=5),  # small[(5, 28-16)], big[(4, 30)]
            Alias(primary=5, alternate=4, primary_height=12),  # small[], big[(4, 30-9)]
            Alias(primary=4, alternate=4, primary_height=21)
        ]
        self.assertEqual(alias_table.to_list(), expected_list)

    def test_alias_table_to_list_does_not_mutate_alias_table(self):
        input_dict = {1: 1}
        alias_table = AliasTable(input_dict)

        alias_list = alias_table.to_list()
        alias_list[0] = Alias(2, 2, 2)
        self.assertNotEqual(alias_table.to_list(), alias_list)

        expected = [Alias(1, 1, 1)]
        self.assertEqual(alias_table.to_list(), expected)



# pylint: disable=missing-docstring, invalid-name, too-many-public-methods
from __future__ import absolute_import

import unittest
import warnings
from itertools import cycle
from sys import version_info

from dicetables.additiveevents import AdditiveEvents
from dicetables.dicerecord import DiceRecord
from dicetables.dicetable import DiceTable, DetailedDiceTable
from dicetables.dieevents import Die
from dicetables.factory.errorhandler import EventsFactoryError
from dicetables.factory.eventsfactory import EventsFactory, Loader, LoaderError
from dicetables.factory.warninghandler import EventsFactoryWarning
from dicetables.tools.dictcombiner import DictCombiner


class NewDiceTableSameInitNoUpdate(DiceTable):
    def __init__(self, event_dic, dice):
        super(NewDiceTableSameInitNoUpdate, self).__init__(event_dic, dice)


class NewDiceTableSameInitUpdate(DiceTable):
    factory_keys = ('get_dict', 'dice_data')

    def __init__(self, event_dic, dice):
        super(NewDiceTableSameInitUpdate, self).__init__(event_dic, dice)


class NewDiceTableNewInitNoUpdate(DiceTable):
    def __init__(self, event_dic, dice, number):
        self.number = number
        super(NewDiceTableNewInitNoUpdate, self).__init__(event_dic, dice)


class NewDiceTableNewInitUpdate(DiceTable):
    factory_keys = ('get_dict', 'dice_data', 'number')
    new_keys = [('number', 0, 'property')]

    def __init__(self, event_dic, dice, number):
        self.number = number
        super(NewDiceTableNewInitUpdate, self).__init__(event_dic, dice)


class TestEventsFactory(unittest.TestCase):
    def setUp(self):
        EventsFactory.reset()

    def assert_EventsFactoryError_contains(self, phrases_tuple, func, *args):
        with self.assertRaises(EventsFactoryError) as cm:
            func(*args)
        error_msg = cm.exception.args[0]
        for phrase in phrases_tuple:
            self.assertIn(phrase, error_msg)

    def assert_EventsFactoryError_code(self, error_code, func, *args):
        self.assert_EventsFactoryError_contains((error_code,), func, *args)

    def assert_EventsFactoryWarning_code(self, code_list, func, *args):
        with warnings.catch_warnings(record=True) as cm:
            warnings.simplefilter("always")
            func(*args)
        self.assertEqual(len(cm), len(code_list))
        for caught in cm:
            self.assertEqual(caught.category, EventsFactoryWarning)
            msg = str(caught.message)
            self.assertTrue(any(code in msg for code in code_list))

    def assert_no_warning(self, func, *args):
        with warnings.catch_warnings(record=True) as cm:
            warnings.simplefilter("always")
            func(*args)
        if len(cm) != 0:
            self.fail('number of warnings: {}'.format(len(cm)))

    def test_assert_EventsFactoryError_contains(self):
        def raise_events_factory_error(msg):
            raise EventsFactoryError(msg)

        self.assert_EventsFactoryError_contains(('a', 'b', 'c'), raise_events_factory_error, 'abcd')

    @unittest.expectedFailure
    def test_assert_EventsFactoryError_contains_fail(self):
        def raise_events_factory_error(msg):
            raise EventsFactoryError(msg)

        self.assert_EventsFactoryError_contains(('a', 'b', 'e'), raise_events_factory_error, 'abcd')

    def test_assert_EventsFactoryError_code(self):
        def raise_events_factory_error(msg):
            raise EventsFactoryError(msg)

        self.assert_EventsFactoryError_code('RAISED', raise_events_factory_error, 'code: RAISED')

    @unittest.expectedFailure
    def test_assert_EventsFactoryError_code_wrong_code(self):
        def raise_events_factory_error(msg):
            raise EventsFactoryError(msg)

        self.assert_EventsFactoryError_code('wrong', raise_events_factory_error, 'code: RAISED')

    def test_assert_EventsFactoryWarning_code(self):
        def func(number):
            warnings.warn('msg\ncode: TEST', EventsFactoryWarning, stacklevel=2)
            return number

        self.assert_EventsFactoryWarning_code(['TEST'], func, 5)

    def test_assert_EventsFactoryWarning_code_multiple(self):
        def func(number):
            warnings.warn('msg\ncode: TEST', EventsFactoryWarning, stacklevel=2)
            warnings.warn('other\ncode: TEST2', EventsFactoryWarning, stacklevel=2)
            return number

        self.assert_EventsFactoryWarning_code(['TEST', 'TEST2'], func, 5)

    @unittest.expectedFailure
    def test_assert_EventsFactoryWarning_code_with_no_warning(self):
        def func(number):
            return number

        self.assert_EventsFactoryWarning_code(['TEST'], func, 5)

    @unittest.expectedFailure
    def test_assert_EventsFactoryWarning_code_with_too_many_warnings(self):
        def func(number):
            warnings.warn('code: TEST', EventsFactoryWarning, stacklevel=2)
            warnings.warn('code: OTHER', Warning, stacklevel=2)
            return number

        self.assert_EventsFactoryWarning_code(['TEST'], func, 5)

    def test_assert_no_warning_with_no_warning(self):
        def func(number):
            return number

        self.assert_no_warning(func, 5)

    @unittest.expectedFailure
    def test_assert_no_warning_with_warning(self):
        def func(number):
            warnings.warn('msg', EventsFactoryWarning, stacklevel=2)
            return number

        self.assert_no_warning(func, 5)

    def test_Loader_no_factory_keys_raises_LoaderError(self):
        self.assertRaises(LoaderError, Loader(EventsFactory).load, NewDiceTableSameInitNoUpdate)

    def test_Loader_bad_factory_keys_raises_EventsFactoryError(self):
        class BadByClassPresentButDifferent(object):
            factory_keys = ('dice_data',)

        EventsFactory.add_class(BadByClassPresentButDifferent, ('get_dict',))

        class BadByBadKeyName(object):
            factory_keys = ('so very very wrong',)

        self.assert_EventsFactoryError_code('CLASS OVERWRITE',
                                            Loader(EventsFactory).load, BadByClassPresentButDifferent)
        self.assert_EventsFactoryError_code('MISSING GETTER',
                                            Loader(EventsFactory).load, BadByBadKeyName)

    def test_Loader_bad_new_keys_raises_EventsFactoryError(self):
        class Bob(object):
            factory_keys = ('dice_data',)
            new_keys = [('dice_data', 'get_dice', [1])]

        self.assert_EventsFactoryError_code('GETTER OVERWRITE', Loader(EventsFactory).load, Bob)

    def test_Loader_class_already_present(self):
        class Bob(object):
            factory_keys = ('dice_data',)

        EventsFactory.add_class(Bob, ('dice_data',))
        self.assertTrue(EventsFactory.has_class(Bob))
        self.assertIsNone(Loader(EventsFactory).load(Bob))
        self.assertTrue(EventsFactory.has_class(Bob))

    def test_Loader_no_new_keys(self):
        class Bob(object):
            factory_keys = ('dice_data',)

        self.assertFalse(EventsFactory.has_class(Bob))
        Loader(EventsFactory).load(Bob)
        self.assertTrue(EventsFactory.has_class(Bob))

    def test_Loader_new_keys(self):
        class Bob(object):
            factory_keys = ('my_method', 'my_property')
            new_keys = [('my_method', 5), ('my_property', 'a', 'property')]

        Loader(EventsFactory).load(Bob)
        self.assertTrue(EventsFactory.get_class_params(Bob), ('my_method', 'my_property'))
        self.assertEqual(EventsFactory.get_getter_string('my_method'), 'method: "my_method", default: 5')
        self.assertEqual(EventsFactory.get_getter_string('my_property'), 'property: "my_property", default: \'a\'')

    def test_EventsFactory_has_class_true(self):
        self.assertTrue(EventsFactory.has_class(AdditiveEvents))

    def test_EventsFactory_has_class_false(self):
        self.assertFalse(EventsFactory.has_class(float))

    def test_EventsFactory_has_getter_true(self):
        self.assertTrue(EventsFactory.has_getter('get_dict'))

    def test_EventsFactory_has_getter_false(self):
        self.assertFalse(EventsFactory.has_getter('not_there'))

    def test_EventsFactory_get_class_params(self):
        self.assertEqual(EventsFactory.get_class_params(AdditiveEvents), ('get_dict',))

    def test_EventsFactory_get_class_params_no_class(self):
        self.assertIsNone(EventsFactory.get_class_params(int))

    def test_EventsFactory_get_getter_string(self):
        get_dict = 'method: "get_dict", default: {0: 1}'
        self.assertEqual(EventsFactory.get_getter_string('get_dict'), get_dict)

    def test_EventsFactory_get_getter_string_no_getters(self):
        self.assertEqual(EventsFactory.get_getter_string('not_there'), 'None')

    def test_EventsFactory_get_keys(self):
        class X(object):
            pass

        classes = ['AdditiveEvents', 'DetailedDiceTable', 'DiceTable']
        getters = ['calc_includes_zeroes', 'dice_data', 'get_dict']
        self.assertEqual(EventsFactory.get_keys(), (classes, getters))

        EventsFactory.add_class(X, ('dice_data',))
        EventsFactory.add_getter('z', 'get', 0)
        classes.append('X')
        getters.append('z')
        self.assertEqual(EventsFactory.get_keys(), (classes, getters))

    def test_EventsFactory_reset(self):
        class Bob(object):
            pass

        EventsFactory.add_class(Bob, ('get_dict',))
        EventsFactory.add_getter('number', 'get_num', 0)
        self.assertTrue(EventsFactory.has_class(Bob))
        self.assertTrue(EventsFactory.has_getter('number'))
        EventsFactory.reset()
        default_classes = ['AdditiveEvents', 'DetailedDiceTable', 'DiceTable']
        default_getters = ['calc_includes_zeroes', 'dice_data', 'get_dict']
        self.assertEqual(EventsFactory.get_keys(), (default_classes, default_getters))

    def test_EventsFactory_add_class(self):
        class Bob(object):
            pass

        self.assertFalse(EventsFactory.has_class(Bob))
        EventsFactory.add_class(Bob, ('get_dict',))
        self.assertTrue(EventsFactory.has_class(Bob))
        self.assertEqual(EventsFactory.get_class_params(Bob), ('get_dict',))

    def test_EventsFactory_add_class_already_has_class_does_not_change(self):
        self.assertTrue(EventsFactory.has_class(AdditiveEvents))
        self.assertEqual(EventsFactory.get_class_params(AdditiveEvents), ('get_dict',))
        EventsFactory.add_class(AdditiveEvents, ('get_dict',))
        self.assertTrue(EventsFactory.has_class(AdditiveEvents))
        self.assertEqual(EventsFactory.get_class_params(AdditiveEvents), ('get_dict',))

    def test_EventsFactory_add_class_already_has_class_raises_EventsFactoryError(self):
        self.assert_EventsFactoryError_code('CLASS OVERWRITE', EventsFactory.add_class, AdditiveEvents, ('dice_data',))

    def test_EventsFactory_add_class_factory_missing_getter_raises_EventsFactoryError(self):
        class Bob(object):
            pass

        self.assert_EventsFactoryError_code('MISSING GETTER', EventsFactory.add_class, Bob, ('not_there',))

    def test_EventsFactory_add_getter_new_method(self):
        EventsFactory.add_getter('get_num', 0)
        self.assertEqual(EventsFactory.get_getter_string('get_num'), 'method: "get_num", default: 0')

    def test_EventsFactory_add_getter_new_property(self):
        EventsFactory.add_getter('get_num', 0, type_str='property')
        self.assertEqual(EventsFactory.get_getter_string('get_num'), 'property: "get_num", default: 0')

    def test_EventsFactory_add_getter_already_there_does_nothing(self):
        EventsFactory.add_getter('get_dict', {0: 1})
        self.assertEqual(EventsFactory.get_getter_string('get_dict'), 'method: "get_dict", default: {0: 1}')

    def test_EventsFactory_add_getter_already_not_equal_raises_error(self):
        self.assert_EventsFactoryError_code('GETTER OVERWRITE',
                                            EventsFactory.add_getter, 'get_dict', {1: 1})
        self.assert_EventsFactoryError_code('GETTER OVERWRITE',
                                            EventsFactory.add_getter, 'get_dict', {0: 1}, 'property')

    def test_EventsFactory_check_no_errors_or_warnings(self):
        self.assert_no_warning(EventsFactory.check, AdditiveEvents)

    def test_EventsFactory_check_no_errors_or_warnings_because_load(self):
        self.assert_no_warning(EventsFactory.check, NewDiceTableSameInitUpdate)

    def test_EventsFactory_check_no_errors_or_warnings_because_load_new_init(self):
        self.assert_no_warning(EventsFactory.check, NewDiceTableNewInitUpdate)

    def test_EventsFactory_check_adds_class_to_factory(self):
        self.assertFalse(EventsFactory.has_class(NewDiceTableNewInitUpdate))
        self.assert_no_warning(EventsFactory.check, NewDiceTableNewInitUpdate)
        self.assertTrue(EventsFactory.has_class(NewDiceTableNewInitUpdate))

    def test_EventsFactory_check_raises_warning(self):
        self.assert_EventsFactoryWarning_code(['CHECK'], EventsFactory.check, NewDiceTableNewInitNoUpdate)

    def test_EventsFactory_check_error(self):
        class Bob(object):
            factory_keys = ('oops',)

        self.assert_EventsFactoryError_code('MISSING GETTER', EventsFactory.check, Bob)

    def test_EventsFactory_check_never_runs_loader_if_class_found(self):
        class Bob(object):
            factory_keys = ('will cause error',)

        EventsFactory.add_class(Bob, ('dice_data',))
        self.assert_no_warning(EventsFactory.check, Bob)

    def test_EventsFactory_check_only_issues_warning_for_totally_unrelated_class(self):
        class Bob(object):
            pass

        self.assert_EventsFactoryWarning_code(['CHECK'], EventsFactory.check, Bob)

    def test_EventFactory_construction__class_in_defaults(self):
        self.assert_no_warning(EventsFactory.new, AdditiveEvents)
        test = EventsFactory.new(AdditiveEvents)
        self.assertIs(type(test), AdditiveEvents)
        self.assertEqual(test.get_dict(), {0: 1})

    def test_EventFactory_construction__new_class_added_and_called(self):
        class Bob(AdditiveEvents):
            pass

        EventsFactory.add_class(Bob, ('get_dict',))
        self.assert_no_warning(EventsFactory.new, Bob)
        test = EventsFactory.new(Bob)
        self.assertIs(type(test), Bob)
        self.assertEqual(test.get_dict(), {0: 1})

    def test_EventsFactory_construction__new_class_has_update_info_init_did_not_change(self):
        self.assert_no_warning(EventsFactory.new, NewDiceTableSameInitUpdate)
        test = EventsFactory.new(NewDiceTableSameInitUpdate)
        self.assertIs(type(test), NewDiceTableSameInitUpdate)
        self.assertEqual(test.get_dict(), {0: 1})
        self.assertEqual(test.get_list(), [])

    def test_EventsFactory_construction__new_class_has_update_info_init_did_change(self):
        self.assert_no_warning(EventsFactory.new, NewDiceTableNewInitUpdate)
        test = EventsFactory.new(NewDiceTableNewInitUpdate)
        self.assertIs(type(test), NewDiceTableNewInitUpdate)
        self.assertEqual(test.get_dict(), {0: 1})
        self.assertEqual(test.get_list(), [])
        self.assertEqual(test.number, 0)

    def test_EventsFactory_construction__new_class_has_no_update_info_init_did_not_change(self):
        self.assert_EventsFactoryWarning_code(['CHECK', 'CONSTRUCT'], EventsFactory.new, NewDiceTableSameInitNoUpdate)
        test = EventsFactory.new(NewDiceTableSameInitNoUpdate)
        self.assertIs(type(test), NewDiceTableSameInitNoUpdate)
        self.assertEqual(test.get_dict(), {0: 1})

    def test_EventsFactory_construction__new_class_has_no_update_info_init_did_change(self):
        self.assert_EventsFactoryWarning_code(['CONSTRUCT'], EventsFactory.new, NewDiceTableNewInitNoUpdate)
        test = EventsFactory.new(NewDiceTableNewInitNoUpdate)
        self.assertIs(type(test), DiceTable)
        self.assertEqual(test.get_dict(), {0: 1})

    def test_EventsFactory_construction__bad_loader_info_ignored_if_class_already_in_factory(self):
        class Bob(object):
            factory_keys = ('whoopsy-doodle',)
            new_keys = [('wrong',), ('very wrong',)]

            def __init__(self, num):
                self.num = num

        EventsFactory.add_getter('num', -5, 'property')
        EventsFactory.add_class(Bob, ('num',))
        self.assert_no_warning(EventsFactory.new, Bob)
        test = EventsFactory.new(Bob)
        self.assertIs(type(test), Bob)
        self.assertEqual(test.num, -5)

    def test_EventsFactory_construction__totally_unrelated_object_with_no_loader_raises_EventsFactoryError(self):
        class Bob(object):
            pass

        self.assert_EventsFactoryError_code('WTF', EventsFactory.new, Bob)

    def test_EventsFactory_construction__new_class_bad_factory_keys_raises_EventsFactoryError(self):
        class BadLoaderKeys(AdditiveEvents):
            factory_keys = ('bad and wrong',)

        self.assert_EventsFactoryError_code('MISSING GETTER', EventsFactory.new, BadLoaderKeys)

    def test_EventsFactory_construction__new_class_bad_new_keys_raises_EventsFactoryError(self):
        class BadLoaderKeys(AdditiveEvents):
            factory_keys = ('dice_data',)
            new_keys = [('dice_data', 5)]

        self.assert_EventsFactoryError_code('GETTER OVERWRITE', EventsFactory.new, BadLoaderKeys)

    def test_EventsFactory_construction__class_present_wrong_signature_wrong_param_EventsFactoryError(self):
        class Bob(AdditiveEvents):
            pass

        EventsFactory.add_class(Bob, ('dice_data',))
        self.assert_EventsFactoryError_code('SIGNATURES DIFFERENT', EventsFactory.new, Bob)

    def test_EventsFactory_construction__class_present_different_signature_too_many_params_EventsFactoryError(self):
        class Bob(AdditiveEvents):
            pass

        EventsFactory.add_class(Bob, ('get_dict', 'dice_data'))
        self.assert_EventsFactoryError_code('SIGNATURES DIFFERENT', EventsFactory.new, Bob)

    def test_EventsFactory_construction__class_present_different_signature_too_few_params_EventsFactoryError(self):
        class Bob(DiceTable):
            pass

        EventsFactory.add_class(Bob, ('get_dict',))
        self.assert_EventsFactoryError_code('SIGNATURES DIFFERENT', EventsFactory.new, Bob)

    def test_EventsFactory_construction__class_present_wrong_signature_wrong_order_EventsFactoryError(self):
        class Bob(DiceTable):
            pass

        EventsFactory.add_class(Bob, ('dice_data', 'get_dict'))
        self.assert_EventsFactoryError_code('SIGNATURES DIFFERENT', EventsFactory.new, Bob)

    def test_EventsFactory_new_AdditiveEvents(self):
        new = EventsFactory.new(AdditiveEvents)
        self.assertIs(type(new), AdditiveEvents)
        self.assertEqual(new.get_dict(), {0: 1})

    def test_EventsFactory_new_DiceTable(self):
        new = EventsFactory.new(DiceTable)
        self.assertIs(type(new), DiceTable)
        self.assertEqual(new.get_dict(), {0: 1})
        self.assertEqual(new.get_list(), [])

    def test_EventsFactory_new_DetailedDiceTable(self):
        new = EventsFactory.new(DetailedDiceTable)
        self.assertIs(type(new), DetailedDiceTable)
        self.assertEqual(new.get_dict(), {0: 1})
        self.assertEqual(new.get_list(), [])
        self.assertTrue(new.calc_includes_zeroes)

    def test_EventsFactory_from_dict_AdditiveEvents(self):
        events = AdditiveEvents.new()
        new_events = EventsFactory.from_dictionary(events, {1: 1})
        self.assertIs(type(new_events), AdditiveEvents)
        self.assertEqual(new_events.get_dict(), {1: 1})

    def test_EventsFactory_from_dict_DiceTable(self):
        events = DiceTable.new()
        new_events = EventsFactory.from_dictionary(events, {1: 1})
        self.assertIs(type(new_events), DiceTable)
        self.assertEqual(new_events.get_dict(), {1: 1})
        self.assertEqual(new_events.get_list(), [])

    def test_EventsFactory_from_dict_DetailedDiceTable(self):
        events = DetailedDiceTable.new()
        new_events = EventsFactory.from_dictionary(events, {1: 1})
        self.assertIs(type(new_events), DetailedDiceTable)
        self.assertEqual(new_events.get_dict(), {1: 1})
        self.assertEqual(new_events.get_list(), [])
        self.assertTrue(new_events.calc_includes_zeroes)

    def test_EventsFactory_from_dict_and_dice_DiceTable(self):
        events = DiceTable.new()
        new_events = EventsFactory.from_dictionary_and_dice(events, {1: 1}, DiceRecord({Die(2): 2}))
        self.assertIs(type(new_events), DiceTable)
        self.assertEqual(new_events.get_dict(), {1: 1})
        self.assertEqual(new_events.get_list(), [(Die(2), 2)])

    def test_EventsFactory_from_dict_and_dice_DetailedDiceTable(self):
        events = DetailedDiceTable({2: 2}, DiceRecord.new(), False)
        new_events = EventsFactory.from_dictionary_and_dice(events, {1: 1}, DiceRecord({Die(2): 2}))
        self.assertIs(type(new_events), DetailedDiceTable)
        self.assertEqual(new_events.get_dict(), {1: 1})
        self.assertEqual(new_events.get_list(), [(Die(2), 2)])
        self.assertFalse(new_events.calc_includes_zeroes)

    def test_EventsFactory_from_params(self):
        start_dict = {2: 2}
        new_dict = {1: 1}

        start_dice = DiceRecord({Die(2): 1})
        new_dice = DiceRecord({Die(3): 3})

        start_num = 5
        new_num = 10

        start = NewDiceTableNewInitUpdate(start_dict, start_dice, start_num)

        new_events1 = EventsFactory.from_params(start, {'number': new_num, 'dice_data': new_dice})
        self.assertEqual(new_events1.get_dict(), start_dict)
        self.assertEqual(new_events1.dice_data(), new_dice)
        self.assertEqual(new_events1.number, new_num)

        new_events2 = EventsFactory.from_params(start, {'get_dict': new_dict})
        self.assertEqual(new_events2.get_dict(), new_dict)
        self.assertEqual(new_events2.dice_data(), start_dice)
        self.assertEqual(new_events2.number, start_num)

        self.assertIs(type(new_events1), NewDiceTableNewInitUpdate)
        self.assertIs(type(new_events2), NewDiceTableNewInitUpdate)

    def test_Events_Factory_warning_correct_message_raised_CHECK(self):
        will_warn_insert = 'TestEventsFactory.test_Events_Factory_warning_correct_message_raised_CHECK.<locals>.'
        if version_info[0] < 3:
            will_warn_insert = ''
        will_warn = 'tests.factory.test_eventsfactory.{}WillWarn'.format(will_warn_insert)
        expected = (
                'factory: <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Warning code: CHECK\n' +
                'Failed to find/add the following class to the EventsFactory - \n' +
                'class: <class \'{}\'>\n'
                '\n' +
                'Warning raised while performing check at instantiation\n' +
                '\n' +
                'SOLUTION:\n' +
                '  class variable: factory_keys = (getter method/property names)\n'
                '  current factory keys are: [\'calc_includes_zeroes\', \'dice_data\', \'get_dict\']\n'
        )

        class WillWarn(DiceTable):
            pass

        self.assert_EventsFactoryWarning_code([expected.format(will_warn)], EventsFactory.check, WillWarn)

    def test_Events_Factory_warning_correct_message_CONSTRUCT(self):
        will_warn_insert = 'TestEventsFactory.test_Events_Factory_warning_correct_message_CONSTRUCT.<locals>.'
        if version_info[0] < 3:
            will_warn_insert = ''
        will_warn = 'tests.factory.test_eventsfactory.{}WillWarn'.format(will_warn_insert)

        expected = (
                'factory: <class \'dicetables.factory.eventsfactory.EventsFactory\'>\n' +
                'Warning code: CONSTRUCT\n' +
                'Failed to find/add the following class to the EventsFactory - \n' +
                'class: <class \'{}\'>\n' +
                '\n' +
                'Class found in factory: <class \'dicetables.dicetable.DiceTable\'>\n'
        )

        class WillWarn(DiceTable):
            pass

        self.assert_EventsFactoryWarning_code([expected.format(will_warn), 'CHECK'], EventsFactory.new, WillWarn)

    def test_Events_Factory_error_correct_message_CLASS_OVERWRITE(self):
        factory_name = '<class \'dicetables.factory.eventsfactory.EventsFactory\'>'
        class_name = '<class \'dicetables.dicetable.DiceTable\'>'
        current_keys = '(\'get_dict\', \'dice_data\')'
        passed_keys = '(\'dice_data\',)'
        expected = (factory_name, class_name, current_keys, passed_keys)
        self.assert_EventsFactoryError_contains(expected, EventsFactory.add_class, DiceTable, ('dice_data',))

    def test_Events_Factory_error_correct_message_GETTER_OVERWRITE(self):
        factory_name = '<class \'dicetables.factory.eventsfactory.EventsFactory\'>'
        getter_key = '\'dice_data\''
        getter_string = 'method: "dice_data", default: DiceRecord({})'
        new_getter_string = 'method: "dice_data", default: [(2, Die(6))]'
        expected = (factory_name, getter_key, getter_string, new_getter_string)

        class Bob(object):
            factory_keys = ('dice_data',)
            new_keys = [('dice_data', [(2, Die(6))], 'method')]

        self.assert_EventsFactoryError_contains(expected, EventsFactory.new, Bob)

    def test_Events_Factory_error_correct_message_MISSING_GETTER(self):
        bob_insert = 'TestEventsFactory.test_Events_Factory_error_correct_message_MISSING_GETTER.<locals>.'
        if version_info[0] < 3:
            bob_insert = ''
        bob = 'tests.factory.test_eventsfactory.{}Bob'.format(bob_insert)
        factory_name = '<class \'dicetables.factory.eventsfactory.EventsFactory\'>'
        factory_keys = '[\'calc_includes_zeroes\', \'dice_data\', \'get_dict\']'
        missing_key = '\'foo\''
        expected = (bob, factory_name, factory_keys, missing_key)

        class Bob(object):
            factory_keys = ('dice_data', 'foo')

        self.assert_EventsFactoryError_contains(expected, EventsFactory.new, Bob)

    def test_Events_Factory_error_correct_message_SIGNATURES_DIFFERENT(self):
        bob_insert = 'TestEventsFactory.test_Events_Factory_error_correct_message_SIGNATURES_DIFFERENT.<locals>.'
        if version_info[0] < 3:
            bob_insert = ''
        bob = 'tests.factory.test_eventsfactory.{}Bob'.format(bob_insert)
        factory_name = '<class \'dicetables.factory.eventsfactory.EventsFactory\'>'
        signature = '(\'get_dict\', \'dice_data\')'
        expected = (bob, factory_name, signature)

        class Bob(object):
            def __init__(self, num):
                self.num = num

        EventsFactory.add_class(Bob, ('get_dict', 'dice_data'))
        self.assert_EventsFactoryError_contains(expected, EventsFactory.new, Bob)

    def test_Events_Factory_error_correct_message_WTF(self):
        bob_insert = 'TestEventsFactory.test_Events_Factory_error_correct_message_WTF.<locals>.'
        if version_info[0] < 3:
            bob_insert = ''
        bob = 'tests.factory.test_eventsfactory.{}Bob'.format(bob_insert)
        factory_name = '<class \'dicetables.factory.eventsfactory.EventsFactory\'>'
        factory_classes = '[\'AdditiveEvents\', \'DetailedDiceTable\', \'DiceTable\']'
        expected = (bob, factory_name, factory_classes)

        class Bob(object):
            pass

        self.assert_EventsFactoryError_contains(expected, EventsFactory.new, Bob)

    def test_BUG_inheritance_issue_RESOLVED_as_best_as_can_be_without_metaclass(self):
        """
        Each class variable of a child class points to its parent class's class variable
        So when parent is re-assigned, child gets re-assigned.  When child get re-assigned, then they no
        longer point to same thing.
        so B(A).  A.num = 5 changes B.num to 5.  B.num = 7 does not change A.num.  A.num = 3 no longer changes B.num
        because it was re-assigned.
        I used StaticDict with EventsFactory so that changes to child dict do not
        mutate parent dict.
        """

        class A(EventsFactory):
            pass

        class B(A):
            pass

        class C(B):
            pass

        self.assertIs(EventsFactory._class_args, C._class_args)
        self.assertIs(B._getters, A._getters)
        B.add_class(int, ('dice_data',))
        self.assertEqual(EventsFactory.get_keys()[0], ['AdditiveEvents', 'DetailedDiceTable', 'DiceTable'])
        self.assertEqual(C.get_keys()[0], ['AdditiveEvents', 'DetailedDiceTable', 'DiceTable', 'int'])
        EventsFactory.reset()
        self.assertIs(EventsFactory._getters, A._getters)
        A.reset()
        self.assertIsNot(EventsFactory._getters, A._getters)
        self.assertEqual(B.get_keys()[0], ['AdditiveEvents', 'DetailedDiceTable', 'DiceTable', 'int'])
        C.reset()
        self.assertIsNot(B._getters, C._getters)

    def test_object_of_astonishing_idiocy(self):
        class DoubleTable(DiceTable):
            """
            TWO TABLES IN ONE!!!
            add_die/remove_die affect primary table.
            combine/remove affect secondary table.
            demonstrates factory method and EventsDictCreator
            """
            factory_keys = ('get_dict', 'get_dict_alt', 'dice_data')
            new_keys = [('get_dict_alt', {0: 1}, 'method')]

            def __init__(self, dict_for_add_die, dict_for_combine, dice_record):
                self._additive = AdditiveEvents(dict_for_combine)
                super(DoubleTable, self).__init__(dict_for_add_die, dice_record)

            def get_dict_alt(self):
                return self._additive.get_dict()

            def combine(self, events, times=1):
                new_alt_dict = DictCombiner(self._additive.get_dict()).combine_by_fastest(events.get_dict(), times)
                return EventsFactory.from_params(self, {'get_dict_alt': new_alt_dict})

            def remove(self, events, times=1):
                new_alt_dict = DictCombiner(self._additive.get_dict()).remove_by_tuple_list(events.get_dict(), times)
                return EventsFactory.from_params(self, {'get_dict_alt': new_alt_dict})

        new = DoubleTable.new()
        d1 = new.add_die(Die(2), 1)
        self.assertEqual(d1.get_dict(), {1: 1, 2: 1})
        self.assertEqual(d1.get_dict_alt(), {0: 1})
        d1_plus = d1.combine(Die(3), 1)
        self.assertEqual(d1_plus.get_dict(), {1: 1, 2: 1})
        self.assertEqual(d1_plus.get_dict_alt(), {1: 1, 2: 1, 3: 1})

    def test_deep_depths_of_dumbness(self):
        class DetachableDiceTable(DiceTable):
            factory_keys = ('get_dict', 'dice_data')
            lyrics = cycle((
                'I woke up this morning with a bad hangover and my {} was missing again.',
                'I can leave it home, when I think it\'s gonna get me in trouble,',
                'I really don\'t like being without my {} for too long. It makes me feel like less of a man,',
                'I saw my {} lying on a blanket next to a broken toaster oven. Some guy was selling it.',
                'I had to buy it off him. He wanted twenty-two bucks, but I talked him down to seventeen.',
                'People sometimes tell me I should get it permanently attached, but I don\'t know.',
                'Even though sometimes it\'s a pain in the ass, I like having a detachable {}.'
            ))

            def __init__(self, events_dict, dice_record):
                super(DetachableDiceTable, self).__init__(events_dict, dice_record)

            def detach(self, die):
                awesome_lyric = next(self.__class__.lyrics).format(die)
                times = self.number_of_dice(die)
                less_of_a_table = self.remove_die(die, times)
                detached = self.new().add_die(die, times)
                return awesome_lyric, detached, less_of_a_table

        new = DetachableDiceTable.new()
        two_d3 = new.add_die(Die(3), 2)
        two_d3_three_d4 = two_d3.add_die(Die(4), 3)
        two_d3_three_d4_four_d5 = two_d3_three_d4.add_die(Die(5), 4)
        msg, detached_1, remainder = two_d3_three_d4_four_d5.detach(Die(5))
        self.assertEqual(msg, 'I woke up this morning with a bad hangover and my D5 was missing again.')
        self.assertIsInstance(detached_1, DetachableDiceTable)
        self.assertEqual(detached_1, DetachableDiceTable.new().add_die(Die(5), 4))
        self.assertEqual(remainder, two_d3_three_d4)
        msg2, empty, detached_2 = detached_1.detach(Die(100))
        self.assertEqual(msg2, 'I can leave it home, when I think it\'s gonna get me in trouble,')
        self.assertEqual(detached_1, detached_2)
        self.assertEqual(str(detached_1), '4D5')
        self.assertEqual(empty.get_dict(), {0: 1})

    def test_a_silly_example_to_show_EventsDictCreator_and_EventsFactory_from_params(self):
        """
        combine events with a single event that occurs once. simply shifts the values of all events.

        AdditiveEvents({1: 2, 3: 4}).combine(AdditiveEvents({10: 1})).get_dict()
        {11: 2, 13: 4}
        AdditiveEvents({1: 2, 3: 4}).combine(AdditiveEvents({100: 1})).get_dict()
        {101: 2, 103: 4}
        """

        class ModifierTable(DiceTable):
            factory_keys = ('get_dict', 'dice_data', 'modifier')
            new_keys = [('modifier', 0, 'property')]

            def __init__(self, events_dict, dice_record, modifier):
                self._mod = modifier
                super(ModifierTable, self).__init__(events_dict, dice_record)

            def __str__(self):
                return '{}\n{:+}'.format(super(ModifierTable, self).__str__(), self._mod)

            @property
            def modifier(self):
                return self._mod

            def change_mod(self, new_modifier):
                change = new_modifier - self._mod
                change_events = AdditiveEvents({change: 1})
                new_dict = DictCombiner(self.get_dict()).combine_by_dictionary(change_events.get_dict(), 1)
                return EventsFactory.from_params(self, {'get_dict': new_dict, 'modifier': new_modifier})

        to_test = ModifierTable.new()
        to_test = to_test.add_die(Die(2), 2)
        self.assertEqual(to_test.get_dict(), {2: 1, 3: 2, 4: 1})
        self.assertEqual(str(to_test), '2D2\n+0')
        to_test = to_test.change_mod(10)
        self.assertEqual(to_test.get_dict(), {12: 1, 13: 2, 14: 1})
        self.assertEqual(str(to_test), '2D2\n+10')
        to_test = to_test.change_mod(-14)
        self.assertEqual(to_test.get_dict(), {-12: 1, -11: 2, -10: 1})
        self.assertEqual(str(to_test), '2D2\n-14')


if __name__ == '__main__':
    unittest.main()
